from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from icloudpd_web.api.session_handler import SessionHandler
from icloudpd_web.api.data_models import AuthenticationResult
from icloudpd_web.api.policy_handler import PolicyStatus
from icloudpd_web.api.logger import build_logger
from icloudpd_web.api.authentication_local import authenticate_secret, save_secret_hash

from dataclasses import dataclass
import socketio
import asyncio
import os

secret_hash_path = os.environ.get("SECRET_HASH_PATH", "~/.icloudpd_web/secret_hash")
secret_hash_path = os.path.abspath(os.path.expanduser(secret_hash_path))
allowed_origins = os.environ.get("ALLOWED_ORIGINS", "*")
allowed_origins = allowed_origins.split(",") if allowed_origins != "*" else "*"


@dataclass
class AppConfig:
    client_ids: set[str]
    allowed_origins: list[str]
    max_sessions: int = os.environ.get("MAX_SESSIONS", 5)
    default_client_id: str = "default-user"
    no_password: bool = os.environ.get("NO_PASSWORD", False)
    always_guest: bool = os.environ.get("ALWAYS_GUEST", False)
    disable_guest: bool = os.environ.get("DISABLE_GUEST", False)
    toml_path: str = os.environ.get("TOML_PATH", "./policies.toml")
    secret_hash_path: str = secret_hash_path


app_config = AppConfig(client_ids=set({"default-user"}), allowed_origins=allowed_origins)


def create_app(
    serve_static: bool = False, static_dir: str = None
) -> tuple[FastAPI, socketio.AsyncServer]:
    # Socket.IO server
    sio = socketio.AsyncServer(
        async_mode="asgi",
        cors_allowed_origins=app_config.allowed_origins,
    )

    print(f"Allowed origins: {app_config.allowed_origins}")

    # FastAPI app
    app = FastAPI(
        title="iCloudPD API", description="API for iCloud Photos Downloader", version="0.1.0"
    )

    # Configure CORS for REST endpoints
    app.add_middleware(
        CORSMiddleware,
        allow_origins=app_config.allowed_origins if app_config.allowed_origins != "*" else ["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    # Mount static files if requested
    if serve_static and static_dir:
        app.mount("/", StaticFiles(directory=static_dir, html=True), name="static")

    # State
    handler_manager: dict[str, SessionHandler] = {}
    # Mapping to track which sids ownership by clientId
    sid_to_client: dict[str, str] = {}

    def find_active_client_id(client_id: str) -> str | None:
        for sid, cid in sid_to_client.items():
            if cid == client_id:
                return sid
        return None

    async def maybe_emit(event: str, client_id: str, preferred_sid: str, *args, **kwargs):
        if preferred_sid in sid_to_client:
            await sio.emit(event, *args, **kwargs, to=preferred_sid)
        elif sid := find_active_client_id(client_id):
            await sio.emit(event, *args, **kwargs, to=sid)
        else:
            print(f"No active session found for client {client_id} when emitting {event}")

    @sio.event
    async def updateAppConfig(sid, key, value):
        try:
            client_id = sid_to_client.get(sid)
            assert client_id, "Client ID not found"
            assert key in {
                "always_guest",
                "disable_guest",
                "no_password",
            }, "Invalid setting to update"
            if (key != "always_guest" or value) and client_id not in app_config.client_ids:
                raise ValueError("Guest user is not allowed to update this setting")

            setattr(app_config, key, value)
            await maybe_emit("app_config_updated", client_id, sid)
        except Exception as e:
            await maybe_emit("error_updating_app_config", client_id, sid, {"error": repr(e)})

    @sio.event
    async def authenticate_local(sid, password):
        if client_id := sid_to_client.get(sid):
            try:
                if authenticate_secret(password, app_config.secret_hash_path):
                    await maybe_emit("server_authenticated", client_id, sid)
                else:
                    await maybe_emit(
                        "server_authentication_failed",
                        client_id,
                        sid,
                        {"error": "Invalid password"},
                    )
            except Exception as e:
                await maybe_emit("server_authentication_failed", client_id, sid, {"error": repr(e)})

    @sio.event
    async def save_secret(sid, old_password, new_password):
        if client_id := sid_to_client.get(sid):
            try:
                if authenticate_secret(old_password, app_config.secret_hash_path):
                    save_secret_hash(new_password, app_config.secret_hash_path)
                    await maybe_emit("server_secret_saved", client_id, sid)
                    await maybe_emit("server_authenticated", client_id, sid)
                else:
                    await maybe_emit(
                        "failed_saving_server_secret",
                        client_id,
                        sid,
                        {"error": "Invalid old password"},
                    )
            except Exception as e:
                await maybe_emit("failed_saving_server_secret", client_id, sid, {"error": repr(e)})

    @sio.event
    async def reset_secret(sid):
        if client_id := sid_to_client.get(sid):
            try:
                print("Resetting server secret, removing all sessions")
                handler_manager.clear()
                try:
                    os.remove(app_config.secret_hash_path)
                except FileNotFoundError:
                    pass
                await maybe_emit("server_secret_reset", client_id, sid)
            except Exception as e:
                await maybe_emit(
                    "failed_resetting_server_secret", client_id, sid, {"error": repr(e)}
                )

    @sio.event
    async def connect(sid, environ, auth):
        """
        Connect a client to the server using clientId for identification.
        """
        # TODO: handle authentication
        client_id = auth.get("clientId", app_config.default_client_id)

        # Store the sid to client mapping
        sid_to_client[sid] = client_id

        if len(sid_to_client) <= app_config.max_sessions:
            if client_id in handler_manager:
                print(f"New session {sid} created for client {client_id}")
            else:
                print(f"New client {client_id} connected with session {sid}")
                handler_manager[client_id] = SessionHandler(
                    saved_policies_path=app_config.toml_path
                )
        else:
            print(f"Disconnecting client {client_id} due to reaching max sessions")
            for sid in sid_to_client.keys():
                if sid_to_client[sid] == client_id:
                    disconnect(sid)

        print(f"Current clients: {list(handler_manager.keys())}")

    @sio.event
    async def disconnect(sid):
        """
        Disconnect and remove handler using sid to client mapping for guest clients.
        """
        if client_id := sid_to_client.pop(sid, None):
            print(f"Client session disconnected: {client_id} (sid: {sid})")
            # Only remove handler if no other sids are using this client_id and the client is a guest
            if client_id not in app_config.client_ids and not any(
                cid == client_id for cid in sid_to_client.values()
            ):
                if client_id in handler_manager:
                    del handler_manager[client_id]
                    print(f"Removed handler for guest {client_id}")

        # print clients and relevant handlers
        for client_id, _ in handler_manager.items():
            print(
                f"Client {client_id} owns sessions {[sid for sid in sid_to_client if sid_to_client[sid] == client_id]}"
            )

    @sio.event
    async def logOut(sid, client_id):
        """
        Log out a client and remove the handler.
        """
        if client_id := sid_to_client.get(sid, None) and client_id in handler_manager:
            del handler_manager[client_id]
            print(f"Removed handler for client {client_id}")

    @sio.event
    async def getServerConfig(sid):
        """
        Get the server config for the client with sid.
        """
        if client_id := sid_to_client.get(sid):
            try:
                viewable_configs = {
                    "always_guest": app_config.always_guest,
                    "disable_guest": app_config.disable_guest,
                    "no_password": app_config.no_password,
                }
                await maybe_emit("server_config", client_id, sid, viewable_configs)
            except Exception as e:
                await maybe_emit("server_config_not_found", client_id, sid, {"error": repr(e)})

    @sio.event
    async def uploadPolicies(sid, toml_content):
        """
        Create policies for the user with sid. Existing policies are replaced.
        """
        if client_id := sid_to_client.get(sid):
            if handler := handler_manager.get(client_id):
                try:
                    handler.replace_policies(toml_content)
                    await maybe_emit("uploaded_policies", client_id, sid, handler.policies)
                except Exception as e:
                    await maybe_emit("error_uploading_policies", client_id, sid, {"error": repr(e)})

    @sio.event
    async def downloadPolicies(sid):
        """
        Download the policies for the user with sid as a TOML string.
        """
        if client_id := sid_to_client.get(sid):
            if handler := handler_manager.get(client_id):
                try:
                    await maybe_emit(
                        "downloaded_policies", client_id, sid, handler.dump_policies_as_toml()
                    )
                except Exception as e:
                    await maybe_emit(
                        "error_downloading_policies", client_id, sid, {"error": repr(e)}
                    )

    @sio.event
    async def getPolicies(sid):
        """
        Get the policies for the user with sid as a list of dictionaries.
        """
        if client_id := sid_to_client.get(sid):
            if handler := handler_manager.get(client_id):
                try:
                    await maybe_emit("policies", client_id, sid, handler.policies)
                except Exception as e:
                    await maybe_emit("internal_error", client_id, sid, {"error": repr(e)})

    @sio.event
    async def savePolicy(sid, policy_name, policy_update):
        """
        Save the policy with the given name and update the parameters. Create a new policy if the name does not exist.
        """
        if client_id := sid_to_client.get(sid):
            if handler := handler_manager.get(client_id):
                try:
                    handler.save_policy(policy_name, **policy_update)
                    await maybe_emit("policies_after_save", client_id, sid, handler.policies)
                except Exception as e:
                    await maybe_emit(
                        "error_saving_policy",
                        client_id,
                        sid,
                        {"policy_name": policy_name, "error": repr(e)},
                    )

    @sio.event
    async def createPolicy(sid, policy):
        """
        Create a new policy with the given name and update the parameters.
        """
        if client_id := sid_to_client.get(sid):
            if handler := handler_manager.get(client_id):
                try:
                    handler.create_policy(**policy)
                    await maybe_emit("policies_after_create", client_id, sid, handler.policies)
                except Exception as e:
                    await maybe_emit(
                        "error_creating_policy",
                        client_id,
                        sid,
                        {"policy_name": policy.get("name", ""), "error": repr(e)},
                    )

    @sio.event
    async def deletePolicy(sid, policy_name):
        """
        Delete a policy with the given name.
        """
        if client_id := sid_to_client.get(sid):
            if handler := handler_manager.get(client_id):
                try:
                    handler.delete_policy(policy_name)
                    await maybe_emit("policies_after_delete", client_id, sid, handler.policies)
                except Exception as e:
                    await maybe_emit(
                        "error_deleting_policy",
                        client_id,
                        sid,
                        {"policy_name": policy_name, "error": repr(e)},
                    )

    @sio.event
    async def authenticate(sid, policy_name, password):
        """
        Authenticate the policy with the given password. Note that this may lead to a MFA request.
        """
        if client_id := sid_to_client.get(sid):
            if handler := handler_manager.get(client_id):
                try:
                    if policy := handler.get_policy(policy_name):
                        result, msg = policy.authenticate(password)
                        match result:
                            case AuthenticationResult.SUCCESS:
                                await maybe_emit(
                                    "authenticated",
                                    client_id,
                                    sid,
                                    {"msg": msg, "policies": handler.policies},
                                )
                            case AuthenticationResult.FAILED:
                                await maybe_emit("authentication_failed", client_id, sid, msg)
                            case AuthenticationResult.MFA_REQUIRED:
                                await maybe_emit("mfa_required", client_id, sid, msg)
                except Exception as e:
                    await maybe_emit("authentication_failed", client_id, sid, repr(e))

    @sio.event
    async def provideMFA(sid, policy_name, mfa_code):
        """
        Finish the authentication for a policy with the MFA code. Note that this may lead to a MFA request if the MFA code is incorrect.
        """
        if client_id := sid_to_client.get(sid):
            if handler := handler_manager.get(client_id):
                try:
                    if policy := handler.get_policy(policy_name):
                        result, msg = policy.provide_mfa(mfa_code)
                        match result:
                            case AuthenticationResult.SUCCESS:
                                await maybe_emit(
                                    "authenticated",
                                    client_id,
                                    sid,
                                    {"msg": msg, "policies": handler.policies},
                                )
                            case AuthenticationResult.MFA_REQUIRED:
                                await maybe_emit("mfa_required", client_id, sid, msg)
                except Exception as e:
                    await maybe_emit("authentication_failed", client_id, sid, repr(e))

    @sio.event
    async def start(sid, policy_name):
        """
        Start the download for the policy with the given name.
        """
        if client_id := sid_to_client.get(sid):
            if handler := handler_manager.get(client_id):
                # Set up logging
                logger, log_capture_stream = build_logger(policy_name)
                try:
                    if policy := handler.get_policy(policy_name):
                        # Check if another policy using the iCloud instance is running
                        if occupying_policy_name := handler.icloud_instance_occupied_by(
                            policy.username
                        ):
                            await maybe_emit(
                                "icloud_is_busy",
                                client_id,
                                sid,
                                f"iCloud user {policy.username} has another policy running: {occupying_policy_name}",
                            )
                            return

                        task = asyncio.create_task(policy.start(logger))
                        last_progress = 0
                        while not task.done():
                            await asyncio.sleep(1)
                            if policy.status == PolicyStatus.RUNNING and (
                                logs := log_capture_stream.read_new_lines()
                                or policy.progress != last_progress
                            ):
                                await maybe_emit(
                                    "download_progress",
                                    client_id,
                                    sid,
                                    {
                                        "policy": policy.dump(),
                                        "logs": logs,
                                    },
                                )
                                last_progress = policy.progress
                        if exception := task.exception():
                            policy.status = PolicyStatus.ERRORED
                            logger.error(f"Download failed: {repr(exception)}")
                            await maybe_emit(
                                "download_failed",
                                client_id,
                                sid,
                                {
                                    "policy": policy.dump(),
                                    "error": repr(exception),
                                    "logs": log_capture_stream.read_new_lines(),
                                },
                            )
                            return

                        await maybe_emit(
                            "download_finished",
                            client_id,
                            sid,
                            {
                                "policy_name": policy_name,
                                "progress": policy.progress,
                                "logs": log_capture_stream.read_new_lines(),
                            },
                        )
                except Exception as e:
                    policy.status = PolicyStatus.ERRORED
                    await maybe_emit(
                        "download_failed",
                        client_id,
                        sid,
                        {
                            "policy": policy.dump(),
                            "error": repr(e),
                            "logs": f"Internal error: {repr(e)}\n",
                        },
                    )
                finally:
                    # Clean up logger and log capture stream
                    if logger and hasattr(logger, "handlers"):
                        for handler in logger.handlers[:]:
                            handler.close()
                            logger.removeHandler(handler)

                    if log_capture_stream and hasattr(log_capture_stream, "close"):
                        log_capture_stream.close()

    @sio.event
    async def interrupt(sid, policy_name):
        """
        Interrupt the download for the policy with the given name.
        """
        if client_id := sid_to_client.get(sid):
            if handler := handler_manager.get(client_id):
                try:
                    if policy := handler.get_policy(policy_name):
                        policy.interrupt()
                except Exception as e:
                    await maybe_emit(
                        "error_interrupting_download",
                        client_id,
                        sid,
                        {"policy_name": policy_name, "error": repr(e)},
                    )

    return app, sio



import uvicorn
import click
import os


@click.command()
@click.option("--host", default="0.0.0.0", help="Host to bind to. Default: 0.0.0.0")
@click.option("--port", default=5000, help="Port to bind to. Default: 5000")
@click.option(
    "--toml-path",
    help="Path to the toml file containing policies definition. The policies will be save to ./policies.toml by default.",
)
@click.option(
    "--allowed-origins",
    multiple=True,
    help="Allowed CORS origins. Pass values with space to specify multiple origins. Warning: '*' will be used if not specified, and this is not recommended for production.",
)
@click.option(
    "--secret-hash-path",
    help="Path to the secret hash file. The secret hash will be saved to ~/.icloudpd_web/secret_hash by default.",
)
@click.option(
    "--max-sessions",
    type=int,
    help="Maximum number of sessions to allow. Default: 5",
)

@click.option(
    "--no-password",
    is_flag=True,
    help="Disable server password authentication. Use it only when access is trusted. This can be configured later in the web interface.",
)

@click.option(
    "--always-guest",
    is_flag=True,
    help="Always login the user as a guest. This can be configured later in the web interface.",
)

@click.option(
    "--disable-guest",
    is_flag=True,
    help="Disable guest login. This can be configured later in the web interface.",
)

# dev options
@click.option(
    "--server-only",
    is_flag=True,
    help="Only run the server without webui. Use it when running the Next.JS server separately.",
)
@click.option(
    "--reload",
    is_flag=True,
    help="Enable auto-reload. This is intended to be used during development.",
)
def main(
    host: str,
    port: int,
    reload: bool,
    server_only: bool,
    toml_path: str,
    allowed_origins: tuple[str],
    secret_hash_path: str,
    max_sessions: int,
    no_password: bool,
    always_guest: bool,
    disable_guest: bool,
):
    """Launch the iCloud Photos Downloader server with the Web interface"""

    if toml_path:
        os.environ["TOML_PATH"] = toml_path

    if allowed_origins:
        os.environ["ALLOWED_ORIGINS"] = ",".join(allowed_origins)

    if secret_hash_path:
        os.environ["SECRET_HASH_PATH"] = secret_hash_path

    if max_sessions:
        os.environ["MAX_SESSIONS"] = str(max_sessions)

    if no_password and always_guest:
        raise click.BadParameter("Cannot enable --no-password and --always-guest together")
    if always_guest and disable_guest:
        raise click.BadParameter("Cannot enable --always-guest and --disable-guest together")

    if no_password:
        os.environ["NO_PASSWORD"] = "true"

    if always_guest:
        os.environ["ALWAYS_GUEST"] = "true"

    if disable_guest:
        os.environ["DISABLE_GUEST"] = "true"

    if server_only:
        uvicorn.run("icloudpd_web.dev:socket_app", host=host, port=port, reload=reload)
    else:
        uvicorn.run("icloudpd_web.main:socket_app", host=host, port=port, reload=reload)


if __name__ == "__main__":
    main()



from typing import Any, Dict, NoReturn, Optional, Sequence
from typing_extensions import override
import typing
import inspect
import json
import logging
from requests import Session

from pyicloud_ipd.exceptions import (
    PyiCloudAPIResponseException,
    PyiCloud2SARequiredException,
    PyiCloudServiceNotActivatedException,
)

LOGGER = logging.getLogger(__name__)

HEADER_DATA = {
    "X-Apple-ID-Account-Country": "account_country",
    "X-Apple-ID-Session-Id": "session_id",
    "X-Apple-Session-Token": "session_token",
    "X-Apple-TwoSV-Trust-Token": "trust_token",
    "X-Apple-TwoSV-Trust-Eligible": "trust_eligible",
    "X-Apple-I-Rscd": "apple_rscd",
    "X-Apple-I-Ercd": "apple_ercd",
    "scnt": "scnt",
}


class PyiCloudPasswordFilter(logging.Filter):
    def __init__(self, password: str):
        super().__init__(password)

    @override
    def filter(self, record: logging.LogRecord) -> bool:
        message = record.getMessage()
        if self.name in message:
            record.msg = message.replace(self.name, "********")
            record.args = [] # type: ignore[assignment] 

        return True


class PyiCloudSession(Session):
    """iCloud session."""

    def __init__(self, service: Any):
        self.service = service
        super().__init__()

    @override
    # type: ignore 
    def request(self, method: str, url, **kwargs):  

        # Charge logging to the right service endpoint
        callee = inspect.stack()[2]
        module = inspect.getmodule(callee[0])
        request_logger = logging.getLogger(module.__name__).getChild("http") #type: ignore[union-attr]
        if self.service.password_filter not in request_logger.filters:
            request_logger.addFilter(self.service.password_filter)

        request_logger.debug("%s %s %s", method, url, kwargs.get("data", ""))

        has_retried = kwargs.get("retried")
        kwargs.pop("retried", None)
        response = super().request(method, url, **kwargs)

        content_type = response.headers.get("Content-Type", "").split(";")[0]
        json_mimetypes = ["application/json", "text/json"]

        request_logger.debug(response.headers)

        for header, value in HEADER_DATA.items():
            if response.headers.get(header):
                session_arg = value
                self.service.session_data.update(
                    {session_arg: response.headers.get(header)}
                )

        # Save session_data to file
        with open(self.service.session_path, "w", encoding="utf-8") as outfile:
            json.dump(self.service.session_data, outfile)
            LOGGER.debug("Saved session data to file")

        # Save cookies to file
        self.cookies.save(ignore_discard=True, ignore_expires=True) # type: ignore[attr-defined]
        LOGGER.debug("Cookies saved to %s", self.service.cookiejar_path)

        if not response.ok and (
            content_type not in json_mimetypes
            or response.status_code in [421, 450, 500]
        ):
            try:
                # pylint: disable=protected-access
                fmip_url = self.service._get_webservice_url("findme")
                if (
                    has_retried is None
                    and response.status_code in [421, 450, 500]
                    and fmip_url in url
                ):
                    # Handle re-authentication for Find My iPhone
                    LOGGER.debug("Re-authenticating Find My iPhone service")
                    try:
                        # If 450, authentication requires a full sign in to the account
                        service = None if response.status_code == 450 else "find"
                        self.service.authenticate(True, service)

                    except PyiCloudAPIResponseException:
                        LOGGER.debug("Re-authentication failed")
                    kwargs["retried"] = True
                    return self.request(method, url, **kwargs)
            except Exception:
                pass

            if has_retried is None and response.status_code in [421, 450, 500]:
                api_error = PyiCloudAPIResponseException(
                    response.reason, str(response.status_code), True
                )
                request_logger.debug(api_error)
                kwargs["retried"] = True
                return self.request(method, url, **kwargs)

            self._raise_error(str(response.status_code), response.reason)

        if content_type not in json_mimetypes:
            if self.service.session_data.get("apple_rscd") == "401":
                code: Optional[str] = "401"
                reason: Optional[str] = "Invalid username/password combination."
                self._raise_error(code or "Unknown", reason or "Unknown")

            return response

        try:
            data = response.json() if response.status_code != 204 else {}
        except:  
            request_logger.warning("Failed to parse response with JSON mimetype")
            return response

        request_logger.debug(data)

        if isinstance(data, dict):
            if data.get("hasError"):
                errors: Optional[Sequence[Dict[str, Any]]] = typing.cast(Optional[Sequence[Dict[str, Any]]], data.get("service_errors"))
                # service_errors returns a list of dict
                #    dict includes the keys: code, title, message, supressDismissal
                # Assuming a single error for now
                # May need to revisit to capture and handle multiple errors
                if errors:
                    code = errors[0].get("code")
                    reason = errors[0].get("message")
                self._raise_error(code or "Unknown", reason or "Unknown")
            elif not data.get("success"):
                reason = data.get("errorMessage")
                reason = reason or data.get("reason")
                reason = reason or data.get("errorReason")
                if not reason and isinstance(data.get("error"), str):
                    reason = data.get("error")
                if not reason and data.get("error"):
                    reason = "Unknown reason"

                code = data.get("errorCode")
                if not code and data.get("serverErrorCode"):
                    code = data.get("serverErrorCode")
                if not code and data.get("error"):
                    code = data.get("error")

                if reason:
                    self._raise_error(code or "Unknown", reason)

        return response

    def _raise_error(self, code: str, reason: str) -> NoReturn:
        if (
            self.service.requires_2sa
            and reason == "Missing X-APPLE-WEBAUTH-TOKEN cookie"
        ):
            raise PyiCloud2SARequiredException(self.service.user["accountName"])
        if code in ("ZONE_NOT_FOUND", "AUTHENTICATION_FAILED"):
            reason = (
                "Please log into https://icloud.com/ to manually "
                "finish setting up your iCloud service"
            )
            api_error: Exception = PyiCloudServiceNotActivatedException(reason, code)
            LOGGER.error(api_error)

            raise (api_error)
        if code == "ACCESS_DENIED":
            reason = (
                reason + ".  Please wait a few minutes then try again."
                "The remote servers might be trying to throttle requests."
            )
        if code in ["421", "450", "500"]:
            reason = "Authentication required for Account."

        api_error = PyiCloudAPIResponseException(reason, code)
        LOGGER.error(api_error)
        raise api_error




from typing import Callable, Dict, Optional, Sequence, Tuple

from icloudpd.mfa_provider import MFAProvider
from pyicloud_ipd.file_match import FileMatchPolicy
from pyicloud_ipd.raw_policy import RawTreatmentPolicy
from pyicloud_ipd.version_size import AssetVersionSize, LivePhotoVersionSize


class Config:
    def __init__(
        self,
        directory: Optional[str],
        username: str,
        auth_only: bool,
        cookie_directory: str,
        primary_sizes: Sequence[AssetVersionSize],
        live_photo_size: LivePhotoVersionSize,
        recent: Optional[int],
        until_found: Optional[int],
        album: str,
        list_albums: bool,
        library: str,
        list_libraries: bool,
        skip_videos: bool,
        skip_live_photos: bool,
        xmp_sidecar: bool,
        force_size: bool,
        auto_delete: bool,
        only_print_filenames: bool,
        folder_structure: str,
        set_exif_datetime: bool,
        smtp_username: Optional[str],
        smtp_host: str,
        smtp_port: int,
        smtp_no_tls: bool,
        notification_email: Optional[str],
        notification_email_from: Optional[str],
        log_level: str,
        no_progress_bar: bool,
        notification_script: Optional[str],
        threads_num: int,
        delete_after_download: bool,
        keep_icloud_recent_days: Optional[int],
        domain: str,
        watch_with_interval: Optional[int],
        dry_run: bool,
        raw_policy: RawTreatmentPolicy,
        password_providers: Dict[
            str, Tuple[Callable[[str], Optional[str]], Callable[[str, str], None]]
        ],
        file_match_policy: FileMatchPolicy,
        mfa_provider: MFAProvider,
        use_os_locale: bool,
    ):
        self.directory = directory
        self.username = username
        self.auth_only = auth_only
        self.cookie_directory = cookie_directory
        self.size = " ".join(str(e) for e in primary_sizes)
        self.live_photo_size = live_photo_size
        self.recent = recent
        self.until_found = until_found
        self.album = album
        self.list_albums = list_albums
        self.library = library
        self.list_libraries = list_libraries
        self.skip_videos = skip_videos
        self.skip_live_photos = skip_live_photos
        self.force_size = force_size
        self.auto_delete = auto_delete
        self.only_print_filenames = only_print_filenames
        self.folder_structure = folder_structure
        self.set_exif_datetime = set_exif_datetime
        self.smtp_username = smtp_username
        self.smtp_host = smtp_host
        self.smtp_port = smtp_port
        self.smtp_no_tls = smtp_no_tls
        self.notification_email = notification_email
        self.notification_email_from = notification_email_from
        self.log_level = log_level
        self.no_progress_bar = no_progress_bar
        self.notification_script = notification_script
        self.threads_num = threads_num
        self.delete_after_download = delete_after_download
        self.keep_icloud_recent_days = keep_icloud_recent_days
        self.domain = domain
        self.watch_with_interval = watch_with_interval
        self.dry_run = dry_run
        self.raw_policy = raw_policy
        self.password_providers = " ".join(str(e) for e in password_providers)
        self.file_match_policy = file_match_policy
        self.mfa_provider = mfa_provider
        self.use_os_locale = use_os_locale



"""Custom logging class and setup function"""

import logging
import sys
from logging import INFO
from typing import Any


class IPDLogger(logging.Logger):
    """Custom logger class with support for tqdm progress bar"""

    def __init__(self, name: str, level: int = INFO):
        logging.Logger.__init__(self, name, level)
        self.tqdm = None

    # If tdqm progress bar is not set, we just write regular log messages
    def set_tqdm(self, tdqm: Any) -> None:
        """Sets the tqdm progress bar"""
        self.tqdm = tdqm

    def set_tqdm_description(self, desc: str, loglevel: int = INFO) -> None:
        """Set tqdm progress bar description, fallback to logging"""
        if self.tqdm is None:
            self.log(loglevel, desc)
        else:
            self.tqdm.set_description(desc)

    def tqdm_write(self, message: str, loglevel: int = INFO) -> None:
        """Write to tqdm progress bar, fallback to logging"""
        if self.tqdm is None:
            self.log(loglevel, message)
        else:
            self.tqdm.write(message)


def setup_logger() -> logging.Logger:
    """Set up logger and add stdout handler"""
    logging.setLoggerClass(IPDLogger)
    logger = logging.getLogger("icloudpd")
    has_stdout_handler = False
    for handler in logger.handlers:
        if handler.name == "stdoutLogger":
            has_stdout_handler = True
    if not has_stdout_handler:
        formatter = logging.Formatter(
            fmt="%(asctime)s %(levelname)-8s %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )
        stdout_handler = logging.StreamHandler(stream=sys.stdout)
        stdout_handler.setFormatter(formatter)
        stdout_handler.name = "stdoutLogger"
        logger.addHandler(stdout_handler)
    return logger



"""Send an email notification when 2SA is expired"""

import datetime
import logging
import smtplib
from typing import Optional, cast


def send_2sa_notification(
    logger: logging.Logger,
    smtp_email: Optional[str],
    smtp_password: Optional[str],
    smtp_host: str,
    smtp_port: int,
    smtp_no_tls: bool,
    to_addr: Optional[str],
    from_addr: Optional[str] = None,
) -> None:
    """Send an email notification when 2SA is expired"""
    to_addr = cast(str, to_addr if to_addr is not None else smtp_email)
    from_addr = (
        from_addr
        if from_addr is not None
        else (f"iCloud Photos Downloader <{smtp_email}>" if smtp_email else to_addr)
    )
    logger.info("Sending 'two-step expired' notification via email...")
    smtp = smtplib.SMTP(smtp_host, smtp_port)
    smtp.set_debuglevel(0)
    # leaving explicit call of connect to not break unit tests, even though it is
    # called implicitly via constructor parameters
    smtp.connect(smtp_host, smtp_port)
    if not smtp_no_tls:
        smtp.starttls()

    if smtp_email is not None and smtp_password is not None:
        smtp.login(smtp_email, smtp_password)

    subj = "icloud_photos_downloader: Two step authentication has expired"
    date = datetime.datetime.now().strftime("%d/%m/%Y %H:%M")

    message_text = """Hello,

Two-step authentication has expired for the icloud_photos_downloader script.
Please log in to your server and run the script manually to update two-step authentication."""

    msg = f"From: {from_addr}\n" + f"To: {to_addr}\nSubject: {subj}\nDate: {date}\n\n{message_text}"

    smtp.sendmail(from_addr, to_addr, msg)
    smtp.quit()




from typing import Optional


class PyiCloudException(Exception):
    """Generic iCloud exception."""
    pass


#API
class PyiCloudAPIResponseException(PyiCloudException):
    """iCloud response exception."""
    def __init__(self, reason:str, code:Optional[str]=None, retry:bool=False):
        self.reason = reason
        self.code = code
        message = reason or ""
        if code:
            message += " (%s)" % code
        if retry:
            message += ". Retrying ..."

        super().__init__(message)


class PyiCloudServiceNotActivatedException(PyiCloudAPIResponseException):
    """iCloud service not activated exception."""
    pass


# Login
class PyiCloudFailedLoginException(PyiCloudException):
    """iCloud failed login exception."""
    pass


class PyiCloud2SARequiredException(PyiCloudException):
    """iCloud 2SA required exception."""
    def __init__(self, apple_id: str):
        message = "Two-step authentication required for account: %s" % apple_id
        super().__init__(message)


class PyiCloudNoStoredPasswordAvailableException(PyiCloudException):
    """iCloud no stored password exception."""
    pass


# Webservice specific
class PyiCloudNoDevicesException(PyiCloudException):
    """iCloud no device exception."""
    pass

# Potentially Deprecated - Further review needed
class PyiCloudConnectionException(PyiCloudException):
    pass


class PyiCloudAPIResponseError(PyiCloudException):
    def __init__(self, reason:str, code:(Optional[int])):
        self.reason = reason
        self.code = code
        message = reason
        if code:
            message += " (%s)" % code

        super(PyiCloudAPIResponseError, self).__init__(message)


class PyiCloud2SARequiredError(PyiCloudException):
    def __init__(self, url: str):
        message = "Two-step authentication required for %s" % url
        super(PyiCloud2SARequiredError, self).__init__(message)


class NoStoredPasswordAvailable(PyiCloudException):
    pass


class PyiCloudServiceNotActivatedErrror(PyiCloudAPIResponseError):
    pass



"""Generate XMP sidecar file from photo asset record"""

from __future__ import annotations

import base64
import json
import logging
import os
import plistlib
import zlib
from datetime import datetime
from typing import Any, NamedTuple
from xml.etree import ElementTree

import dateutil.tz

from foundation import version_info

exif_tool = None


class XMPMetadata(NamedTuple):
    XMPToolkit: str
    Title: str | None
    Description: str | None
    Orientation: int | None
    Make: str | None
    DigitalSourceType: str | None
    Keywords: list[str] | None
    GPSAltitude: float | None
    GPSLatitude: float | None
    GPSLongitude: float | None
    GPSSpeed: float | None
    GPSTimeStamp: datetime | None
    CreateDate: datetime | None
    Rating: int | None


def generate_xmp_file(
    logger: logging.Logger, download_path: str, asset_record: dict[str, Any]
) -> None:
    sidecar_path: str = download_path + ".xmp"
    can_write_file: bool = True
    if os.path.exists(sidecar_path) and os.path.getsize(sidecar_path) != 0:
        can_write_file = False
        try:
            root = ElementTree.parse(sidecar_path).getroot()
            xmptk_value = root.attrib.get("{adobe:ns:meta/}xmptk")
            if not xmptk_value or not xmptk_value.startswith("icloudpd"):
                logger.info(f"Not overwriting XMP file {sidecar_path} created by {xmptk_value}")
            else:
                can_write_file = True
        except ElementTree.ParseError as e:
            logger.info(f"Not overwriting XMP file {sidecar_path} due to parser error: {e}")

    # decode asset record fields
    # for k in asset_record['fields']:
    #    if asset_record["fields"][k]['type'] == "ENCRYPTED_BYTES":
    # try:
    #     asset_record["fields"][k]['decoded'] = plistlib.loads(base64.b64decode(asset_record['fields'][k]['value']), fmt=plistlib.FMT_BINARY)
    # except plistlib.InvalidFileException:
    #     try:
    #         asset_record["fields"][k]['decoded'] =  json.loads(zlib.decompress(base64.b64decode(asset_record['fields'][k]['value']),-zlib.MAX_WBITS))
    #     except Exception as e:
    #         asset_record["fields"][k]['decoded'] = base64.b64decode(asset_record['fields'][k]['value']).decode("utf-8")
    # json.dump(asset_record["fields"],         open(download_path + ".ar.json", "w"),         indent=4,        default=str,        sort_keys=True)

    if can_write_file:
        xmp_metadata: XMPMetadata = build_metadata(asset_record)
        xml_doc: ElementTree.Element = generate_xml(xmp_metadata)
        # Write the XML to the file
        with open(sidecar_path, "wb") as f:
            f.write(ElementTree.tostring(xml_doc, encoding="utf-8", xml_declaration=True))


def build_metadata(asset_record: dict[str, Any]) -> XMPMetadata:
    """Build XMP metadata from asset record"""

    title = None
    if "captionEnc" in asset_record["fields"]:
        title = base64.b64decode(asset_record["fields"]["captionEnc"]["value"]).decode("utf-8")

    description = None
    if "extendedDescEnc" in asset_record["fields"]:
        description = base64.b64decode(asset_record["fields"]["extendedDescEnc"]["value"]).decode(
            "utf-8"
        )

    orientation = None
    if "adjustmentSimpleDataEnc" in asset_record["fields"]:
        adjustments = json.loads(
            zlib.decompress(
                base64.b64decode(asset_record["fields"]["adjustmentSimpleDataEnc"]["value"]),
                -zlib.MAX_WBITS,
            )
        )
        if "metadata" in adjustments and "orientation" in adjustments["metadata"]:
            orientation = adjustments["metadata"]["orientation"]

    make, digital_source_type = None, None
    if (
        "assetSubtypeV2" in asset_record["fields"]
        and int(asset_record["fields"]["assetSubtypeV2"]["value"]) == 3
    ):
        make = "Screenshot"
        digital_source_type = "screenCapture"

    keywords = None
    if "keywordsEnc" in asset_record["fields"] and len(asset_record["fields"]["keywordsEnc"]) > 0:
        keywords = plistlib.loads(
            base64.b64decode(asset_record["fields"]["keywordsEnc"]["value"]),
            fmt=plistlib.FMT_BINARY,
        )

    if "locationEnc" in asset_record["fields"]:
        location = plistlib.loads(
            base64.b64decode(asset_record["fields"]["locationEnc"]["value"]),
            fmt=plistlib.FMT_BINARY,
        )
        gps_altitude = location.get("alt")
        gps_latitude = location.get("lat")
        gps_longitude = location.get("lon")
        gps_speed = location.get("speed")
        gps_timestamp = (
            location.get("timestamp") if isinstance(location.get("timestamp"), datetime) else None
        )
    else:
        gps_altitude, gps_latitude, gps_longitude, gps_speed, gps_timestamp = (
            None,
            None,
            None,
            None,
            None,
        )

    create_date = None
    if "assetDate" in asset_record["fields"]:
        timezone_offset = 0
        if "timeZoneOffset" in asset_record["fields"]:
            timezone_offset = asset_record["fields"]["timeZoneOffset"]["value"]
        create_date = datetime.fromtimestamp(
            int(asset_record["fields"]["assetDate"]["value"]) / 1000,
            tz=dateutil.tz.tzoffset(None, timezone_offset),
        )

    rating = None
    # Hidden or Deleted Photos should be marked as rejected (needs running as --album "Hidden" or --album "Recently Deleted")
    if (
        "isHidden" in asset_record["fields"] and asset_record["fields"]["isHidden"]["value"] == 1
    ) or (
        "isDeleted" in asset_record["fields"] and asset_record["fields"]["isDeleted"]["value"] == 1
    ):
        rating = -1  # -1 means rejected: https://www.iptc.org/std/photometadata/specification/IPTC-PhotoMetadata#image-rating
    # only mark photo as favorite if not hidden or deleted
    elif asset_record["fields"]["isFavorite"]["value"] == 1:
        rating = 5

    return XMPMetadata(
        XMPToolkit="icloudpd " + version_info.version + "+" + version_info.commit_sha,
        Title=title,
        Description=description,
        Orientation=orientation,
        Make=make,
        DigitalSourceType=digital_source_type,
        Keywords=keywords,
        GPSAltitude=gps_altitude,
        GPSLatitude=gps_latitude,
        GPSLongitude=gps_longitude,
        GPSSpeed=gps_speed,
        GPSTimeStamp=gps_timestamp,
        CreateDate=create_date,
        Rating=rating,
    )


def generate_xml(metadata: XMPMetadata) -> ElementTree.Element:
    # Create the root element
    xml_doc = ElementTree.Element(
        "x:xml_doc", {"xmlns:x": "adobe:ns:meta/", "x:xmptk": metadata.XMPToolkit}
    )

    # Create the RDF element
    rdf = ElementTree.SubElement(
        xml_doc, "rdf:RDF", {"xmlns:rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#"}
    )

    # Create the Description elements
    description_dc = ElementTree.Element(
        "rdf:Description",
        {
            "rdf:about": "",
            "xmlns:dc": "http://purl.org/dc/elements/1.1/",
        },
    )
    description_exif = ElementTree.Element(
        "rdf:Description",
        {
            "rdf:about": "",
            "xmlns:exif": "http://ns.adobe.com/exif/1.0/",
        },
    )
    description_iptc4xmpext = ElementTree.Element(
        "rdf:Description",
        {
            "rdf:about": "",
            "xmlns:Iptc4xmpExt": "http://iptc.org/std/Iptc4xmpExt/2008-02-29/",
        },
    )
    description_photoshop = ElementTree.Element(
        "rdf:Description",
        {
            "rdf:about": "",
            "xmlns:photoshop": "http://ns.adobe.com/photoshop/1.0/",
        },
    )
    description_tiff = ElementTree.Element(
        "rdf:Description",
        {
            "rdf:about": "",
            "xmlns:tiff": "http://ns.adobe.com/tiff/1.0/",
        },
    )
    description_xmp = ElementTree.Element(
        "rdf:Description",
        {
            "rdf:about": "",
            "xmlns:xmp": "http://ns.adobe.com/xap/1.0/",
        },
    )
    if metadata.Title:
        ElementTree.SubElement(description_dc, "dc:title").text = metadata.Title
    if metadata.Description:
        ElementTree.SubElement(description_dc, "dc:description").text = metadata.Description

    if metadata.Orientation:
        ElementTree.SubElement(description_tiff, "tiff:Orientation").text = str(
            metadata.Orientation
        )
    if metadata.Make:
        ElementTree.SubElement(description_tiff, "tiff:Make").text = metadata.Make
    if metadata.DigitalSourceType:
        ElementTree.SubElement(
            description_iptc4xmpext, "Iptc4xmpExt:DigitalSourceType"
        ).text = metadata.DigitalSourceType

    if metadata.Keywords:
        subject = ElementTree.SubElement(description_dc, "dc:subject")
        seq = ElementTree.SubElement(subject, "rdf:Seq")
        for keyword in metadata.Keywords:
            ElementTree.SubElement(seq, "rdf:li").text = keyword

    if metadata.GPSAltitude:
        ElementTree.SubElement(description_exif, "exif:GPSAltitude").text = str(
            metadata.GPSAltitude
        )
    if metadata.GPSLatitude:
        ElementTree.SubElement(description_exif, "exif:GPSLatitude").text = str(
            metadata.GPSLatitude
        )
    if metadata.GPSLongitude:
        ElementTree.SubElement(description_exif, "exif:GPSLongitude").text = str(
            metadata.GPSLongitude
        )
    if metadata.GPSSpeed:
        ElementTree.SubElement(description_exif, "exif:GPSSpeed").text = str(metadata.GPSSpeed)
    if metadata.GPSTimeStamp:
        ElementTree.SubElement(
            description_exif, "exif:GPSTimeStamp"
        ).text = metadata.GPSTimeStamp.strftime("%Y-%m-%dT%H:%M:%S%z")

    if metadata.CreateDate:
        ElementTree.SubElement(
            description_xmp, "xmp:CreateDate"
        ).text = metadata.CreateDate.strftime("%Y-%m-%dT%H:%M:%S%z")
        ElementTree.SubElement(
            description_photoshop, "photoshop:DateCreated"
        ).text = metadata.CreateDate.strftime(
            "%Y-%m-%dT%H:%M:%S%z"
        )  # Apple Photos uses this field when exporting an XMP sidecar

    if metadata.Rating:
        ElementTree.SubElement(description_xmp, "xmp:Rating").text = str(metadata.Rating)
    if len(list(description_dc)) > 0:
        rdf.append(description_dc)
    if len(list(description_exif)) > 0:
        rdf.append(description_exif)
    if len(list(description_iptc4xmpext)) > 0:
        rdf.append(description_iptc4xmpext)
    if len(list(description_photoshop)) > 0:
        rdf.append(description_photoshop)
    if len(list(description_tiff)) > 0:
        rdf.append(description_tiff)
    if len(list(description_xmp)) > 0:
        rdf.append(description_xmp)

    return xml_doc



#!/usr/bin/env python
"""Main script that uses Click to parse command-line arguments"""

from multiprocessing import freeze_support

import foundation
from foundation.core import compose, constant, identity
from icloudpd.mfa_provider import MFAProvider
from pyicloud_ipd.item_type import AssetItemType  # fmt: skip

freeze_support()  # fmt: skip # fixing tqdm on macos

import datetime
import itertools
import json
import logging
import os
import subprocess
import sys
import time
import typing
import urllib
from functools import partial
from logging import Logger
from threading import Thread
from typing import (
    Callable,
    Dict,
    Iterable,
    NoReturn,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    cast,
)

import click
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm
from tzlocal import get_localzone

from icloudpd import constants, download, exif_datetime
from icloudpd.authentication import TwoStepAuthRequiredError, authenticator
from icloudpd.autodelete import autodelete_photos
from icloudpd.config import Config
from icloudpd.counter import Counter
from icloudpd.email_notifications import send_2sa_notification
from icloudpd.paths import clean_filename, local_download_path, remove_unicode_chars
from icloudpd.server import serve_app
from icloudpd.status import Status, StatusExchange
from icloudpd.string_helpers import truncate_middle
from icloudpd.xmp_sidecar import generate_xmp_file
from pyicloud_ipd.base import PyiCloudService
from pyicloud_ipd.exceptions import PyiCloudAPIResponseException
from pyicloud_ipd.file_match import FileMatchPolicy
from pyicloud_ipd.raw_policy import RawTreatmentPolicy
from pyicloud_ipd.services.photos import PhotoAsset, PhotoLibrary, PhotosService
from pyicloud_ipd.utils import (
    add_suffix_to_filename,
    disambiguate_filenames,
    get_password_from_keyring,
    size_to_suffix,
    store_password_in_keyring,
)
from pyicloud_ipd.version_size import AssetVersionSize, LivePhotoVersionSize


def build_filename_cleaner(
    _ctx: click.Context, _param: click.Parameter, is_keep_unicode: bool
) -> Callable[[str], str]:
    """Map keep_unicode parameter for function for cleaning filenames"""
    # redefining typed vars instead of using in ternary directly is a mypy hack
    r: Callable[[str], str] = remove_unicode_chars
    i: Callable[[str], str] = identity
    return compose(
        (r if not is_keep_unicode else i),
        clean_filename,
    )


def lp_filename_concatinator(filename: str) -> str:
    name, ext = os.path.splitext(filename)
    if not ext:
        return filename
    return name + ("_HEVC.MOV" if ext.lower().endswith(".heic") else ".MOV")


def lp_filename_original(filename: str) -> str:
    name, ext = os.path.splitext(filename)
    if not ext:
        return filename
    return name + ".MOV"


def build_lp_filename_generator(
    _ctx: click.Context, _param: click.Parameter, lp_filename_policy: str
) -> Callable[[str], str]:
    # redefining typed vars instead of using in ternary directly is a mypy hack
    return lp_filename_original if lp_filename_policy == "original" else lp_filename_concatinator


def raw_policy_generator(
    _ctx: click.Context, _param: click.Parameter, raw_policy: str
) -> RawTreatmentPolicy:
    # redefining typed vars instead of using in ternary directly is a mypy hack
    if raw_policy == "as-is":
        return RawTreatmentPolicy.AS_IS
    elif raw_policy == "original":
        return RawTreatmentPolicy.AS_ORIGINAL
    elif raw_policy == "alternative":
        return RawTreatmentPolicy.AS_ALTERNATIVE
    else:
        raise ValueError(f"policy was provided with unsupported value of '{raw_policy}'")


def size_generator(
    _ctx: click.Context, _param: click.Parameter, sizes: Sequence[str]
) -> Sequence[AssetVersionSize]:
    def _map(size: str) -> AssetVersionSize:
        if size == "original":
            return AssetVersionSize.ORIGINAL
        elif size == "adjusted":
            return AssetVersionSize.ADJUSTED
        elif size == "alternative":
            return AssetVersionSize.ALTERNATIVE
        elif size == "medium":
            return AssetVersionSize.MEDIUM
        elif size == "thumb":
            return AssetVersionSize.THUMB
        else:
            raise ValueError(f"size was provided with unsupported value of '{size}'")

    return [_map(_s) for _s in sizes]


def mfa_provider_generator(
    _ctx: click.Context, _param: click.Parameter, provider: str
) -> MFAProvider:
    if provider == "console":
        return MFAProvider.CONSOLE
    elif provider == "webui":
        return MFAProvider.WEBUI
    else:
        raise ValueError(f"mfa provider has unsupported value of '{provider}'")


def ask_password_in_console(_user: str) -> Optional[str]:
    return typing.cast(Optional[str], click.prompt("iCloud Password", hide_input=True))
    # return getpass.getpass(
    #         f'iCloud Password for {_user}:'
    #     )


def get_password_from_webui(
    logger: Logger, status_exchange: StatusExchange
) -> Callable[[str], Optional[str]]:
    def _intern(_user: str) -> Optional[str]:
        """Request two-factor authentication through Webui."""
        if not status_exchange.replace_status(Status.NO_INPUT_NEEDED, Status.NEED_PASSWORD):
            logger.error("Expected NO_INPUT_NEEDED, but got something else")
            return None

        # wait for input
        while True:
            status = status_exchange.get_status()
            if status == Status.NEED_PASSWORD:
                time.sleep(1)
            else:
                break
        if status_exchange.replace_status(Status.SUPPLIED_PASSWORD, Status.CHECKING_PASSWORD):
            password = status_exchange.get_payload()
            if not password:
                logger.error("Internal error: did not get password for SUPPLIED_PASSWORD status")
                status_exchange.replace_status(
                    Status.CHECKING_PASSWORD, Status.NO_INPUT_NEEDED
                )  # TODO Error
                return None
            return password

        return None  # TODO

    return _intern


def update_password_status_in_webui(status_exchange: StatusExchange) -> Callable[[str, str], None]:
    def _intern(_u: str, _p: str) -> None:
        # TODO we are not handling wrong passwords...
        status_exchange.replace_status(Status.CHECKING_PASSWORD, Status.NO_INPUT_NEEDED)
        return None

    return _intern


# def get_click_param_by_name(_name: str, _params: List[Parameter]) -> Optional[Parameter]:
#     _with_password = [_p for _p in _params if _name in _p.name]
#     if len(_with_password) == 0:
#         return None
#     return _with_password[0]


def dummy_password_writter(_u: str, _p: str) -> None:
    pass


def keyring_password_writter(logger: Logger) -> Callable[[str, str], None]:
    def _intern(username: str, password: str) -> None:
        try:
            store_password_in_keyring(username, password)
        except Exception:
            logger.warning("Password was not saved to keyring")

    return _intern


def password_provider_generator(
    _ctx: click.Context, _param: click.Parameter, providers: Sequence[str]
) -> Dict[str, Tuple[Callable[[str], Optional[str]], Callable[[str, str], None]]]:
    def _map(provider: str) -> Tuple[Callable[[str], Optional[str]], Callable[[str, str], None]]:
        if provider == "webui":
            return (ask_password_in_console, dummy_password_writter)
        if provider == "console":
            return (ask_password_in_console, dummy_password_writter)
        elif provider == "keyring":
            return (get_password_from_keyring, dummy_password_writter)
        elif provider == "parameter":
            # TODO get from parameter
            # _param: Optional[Parameter] = get_click_param_by_name("password", _ctx.command.params)
            # if _param:
            #     _password: str = _param.consume_value(_ctx, {})
            #     return constant(_password)
            return (constant(None), dummy_password_writter)
        else:
            raise ValueError(f"password provider was given an unsupported value of '{provider}'")

    return dict([(_s, _map(_s)) for _s in providers])


def lp_size_generator(
    _ctx: click.Context, _param: click.Parameter, size: str
) -> LivePhotoVersionSize:
    if size == "original":
        return LivePhotoVersionSize.ORIGINAL
    elif size == "medium":
        return LivePhotoVersionSize.MEDIUM
    elif size == "thumb":
        return LivePhotoVersionSize.THUMB
    else:
        raise ValueError(f"size was provided with unsupported value of '{size}'")


def file_match_policy_generator(
    _ctx: click.Context, _param: click.Parameter, policy: str
) -> FileMatchPolicy:
    if policy == "name-size-dedup-with-suffix":
        return FileMatchPolicy.NAME_SIZE_DEDUP_WITH_SUFFIX
    elif policy == "name-id7":
        return FileMatchPolicy.NAME_ID7
    else:
        raise ValueError(f"policy was provided with unsupported value of '{policy}'")


def locale_setter(_ctx: click.Context, _param: click.Parameter, use_os_locale: bool) -> bool:
    # set locale
    if use_os_locale:
        from locale import LC_ALL, setlocale

        setlocale(LC_ALL, "")
    return use_os_locale


def report_version(ctx: click.Context, _param: click.Parameter, value: bool) -> bool:
    if not value:
        return value
    vi = foundation.version_info_formatted()
    click.echo(vi)
    ctx.exit()


# Must import the constants object so that we can mock values in tests.

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


@click.command(context_settings=CONTEXT_SETTINGS, options_metavar="<options>", no_args_is_help=True)
@click.option(
    "-d",
    "--directory",
    help="Local directory that should be used for download",
    type=click.Path(exists=True),
    metavar="<directory>",
)
@click.option(
    "-u",
    "--username",
    help="Your iCloud username or email address",
    metavar="<username>",
    prompt="iCloud username/email",
)
@click.option(
    "-p",
    "--password",
    help="Your iCloud password " "(default: use PyiCloud keyring or prompt for password)",
    metavar="<password>",
    # is_eager=True,
)
@click.option(
    "--auth-only",
    help="Create/Update cookie and session tokens only.",
    is_flag=True,
)
@click.option(
    "--cookie-directory",
    help="Directory to store cookies for authentication " "(default: ~/.pyicloud)",
    metavar="</cookie/directory>",
    default="~/.pyicloud",
)
@click.option(
    "--size",
    help="Image size to download. `medium` and `thumb` will always be added as suffixes to filenames, `adjusted` and `alternative` only if conflicting, `original` - never. If `adjusted` or `alternative` specified and is missing, then `original` is used.",
    type=click.Choice(["original", "medium", "thumb", "adjusted", "alternative"]),
    default=["original"],
    multiple=True,
    show_default=True,
    callback=size_generator,
)
@click.option(
    "--live-photo-size",
    help="Live Photo video size to download",
    type=click.Choice(["original", "medium", "thumb"]),
    default="original",
    show_default=True,
    callback=lp_size_generator,
)
@click.option(
    "--recent",
    help="Number of recent photos to download (default: download all photos)",
    type=click.IntRange(0),
)
@click.option(
    "--until-found",
    help="Download most recently added photos until we find x number of "
    "previously downloaded consecutive photos (default: download all photos)",
    type=click.IntRange(0),
)
@click.option(
    "-a",
    "--album",
    help="Album to download (default: All Photos)",
    metavar="<album>",
    default="All Photos",
)
@click.option(
    "-l",
    "--list-albums",
    help="Lists the available albums",
    is_flag=True,
)
@click.option(
    "--library",
    help="Library to download (default: Personal Library)",
    metavar="<library>",
    default="PrimarySync",
)
@click.option(
    "--list-libraries",
    help="Lists the available libraries",
    is_flag=True,
)
@click.option(
    "--skip-videos",
    help="Don't download any videos (default: Download all photos and videos)",
    is_flag=True,
)
@click.option(
    "--skip-live-photos",
    help="Don't download any live photos (default: Download live photos)",
    is_flag=True,
)
@click.option(
    "--xmp-sidecar",
    help="Export additional data as XMP sidecar files (default: don't export)",
    is_flag=True,
)
@click.option(
    "--force-size",
    help="Only download the requested size (`adjusted` and `alternate` will not be forced)"
    + "(default: download original if size is not available)",
    is_flag=True,
)
@click.option(
    "--auto-delete",
    help='Scans the "Recently Deleted" folder and deletes any files found in there. '
    + "(If you restore the photo in iCloud, it will be downloaded again.)",
    is_flag=True,
)
@click.option(
    "--only-print-filenames",
    help="Only prints the filenames of all files that will be downloaded "
    "(not including files that are already downloaded.)"
    + "(Does not download or delete any files.)",
    is_flag=True,
)
@click.option(
    "--folder-structure",
    help="Folder structure (default: {:%Y/%m/%d}). "
    "If set to 'none' all photos will just be placed into the download directory",
    metavar="<folder_structure>",
    default="{:%Y/%m/%d}",
)
@click.option(
    "--set-exif-datetime",
    help="Write the DateTimeOriginal exif tag from file creation date, " + "if it doesn't exist.",
    is_flag=True,
)
@click.option(
    "--smtp-username",
    help="Your SMTP username, for sending email notifications when "
    "two-step authentication expires.",
    metavar="<smtp_username>",
)
@click.option(
    "--smtp-password",
    help="Your SMTP password, for sending email notifications when "
    "two-step authentication expires.",
    metavar="<smtp_password>",
)
@click.option(
    "--smtp-host",
    help="Your SMTP server host. Defaults to: smtp.gmail.com",
    metavar="<smtp_host>",
    default="smtp.gmail.com",
)
@click.option(
    "--smtp-port",
    help="Your SMTP server port. Default: 587 (Gmail)",
    metavar="<smtp_port>",
    type=click.IntRange(0),
    default=587,
)
@click.option(
    "--smtp-no-tls",
    help="Pass this flag to disable TLS for SMTP (TLS is required for Gmail)",
    metavar="<smtp_no_tls>",
    is_flag=True,
)
@click.option(
    "--notification-email",
    help="Email address where you would like to receive email notifications. "
    "Default: SMTP username",
    metavar="<notification_email>",
)
@click.option(
    "--notification-email-from",
    help="Email address from which you would like to receive email notifications. "
    "Default: SMTP username or notification-email",
    metavar="<notification_email_from>",
)
@click.option(
    "--notification-script",
    type=click.Path(),
    help="Runs an external script when two factor authentication expires. "
    "(path required: /path/to/my/script.sh)",
)
@click.option(
    "--log-level",
    help="Log level (default: debug)",
    type=click.Choice(["debug", "info", "error"]),
    default="debug",
)
@click.option(
    "--no-progress-bar",
    help="Disables the one-line progress bar and prints log messages on separate lines "
    "(Progress bar is disabled by default if there is no tty attached)",
    is_flag=True,
)
@click.option(
    "--threads-num",
    help="Number of cpu threads - deprecated & always 1. To be removed in future version",
    type=click.IntRange(1),
    default=1,
)
@click.option(
    "--delete-after-download",
    help="Delete the photo/video after download it."
    + ' The deleted items will be appear in the "Recently Deleted".'
    + " Therefore, should not combine with --auto-delete option.",
    is_flag=True,
)
@click.option(
    "--keep-icloud-recent-days",
    help="Keep photos newer than this many days in iCloud. Deletes the rest. "
    + "If set to 0, all photos will be deleted from iCloud.",
    type=click.IntRange(0),
    default=None,
)
@click.option(
    "--domain",
    help="What iCloud root domain to use. Use 'cn' for mainland China (default: 'com')",
    type=click.Choice(["com", "cn"]),
    default="com",
)
@click.option(
    "--watch-with-interval",
    help="Run downloading in a infinite cycle, waiting specified seconds between runs",
    type=click.IntRange(1),
)
@click.option(
    "--dry-run",
    help="Do not modify local system or iCloud",
    is_flag=True,
    default=False,
)
@click.option(
    "--keep-unicode-in-filenames",
    "filename_cleaner",
    help="Keep unicode chars in file names or remove non all ascii chars",
    type=bool,
    default=False,
    callback=build_filename_cleaner,
)
@click.option(
    "--live-photo-mov-filename-policy",
    "lp_filename_generator",
    help="How to produce filenames for video portion of live photos: `suffix` will add _HEVC suffix and `original` will keep filename as it is.",
    type=click.Choice(["suffix", "original"], case_sensitive=False),
    default="suffix",
    callback=build_lp_filename_generator,
)
@click.option(
    "--align-raw",
    "raw_policy",
    help="For photo assets with raw and jpeg, treat raw always in the specified size: `original` (raw+jpeg), `alternative` (jpeg+raw), or unchanged (as-is). It matters when choosing sizes to download",
    type=click.Choice(["as-is", "original", "alternative"], case_sensitive=False),
    default="as-is",
    show_default=True,
    callback=raw_policy_generator,
)
@click.option(
    "--password-provider",
    "password_providers",
    help="Specifies passwords provider to check in the specified order",
    type=click.Choice(["console", "keyring", "parameter", "webui"], case_sensitive=False),
    default=["parameter", "keyring", "console"],
    show_default=True,
    multiple=True,
    callback=password_provider_generator,
)
@click.option(
    "--file-match-policy",
    "file_match_policy",
    help="Policy to identify existing files and de-duplicate. `name-size-dedup-with-suffix` appends file size to deduplicate. `name-id7` adds asset id from iCloud to all file names and does not de-duplicate.",
    type=click.Choice(["name-size-dedup-with-suffix", "name-id7"], case_sensitive=False),
    default="name-size-dedup-with-suffix",
    show_default=True,
    callback=file_match_policy_generator,
)
@click.option(
    "--mfa-provider",
    help="Specified where to get MFA code from",
    type=click.Choice(["console", "webui"], case_sensitive=False),
    default="console",
    show_default=True,
    callback=mfa_provider_generator,
)
@click.option(
    "--use-os-locale",
    help="Use locale of the host OS to format dates",
    is_flag=True,
    default=False,
    is_eager=True,
    callback=locale_setter,
)
@click.option(
    "--version",
    help="Show the version, commit hash and timestamp",
    is_flag=True,
    expose_value=False,
    is_eager=True,
    callback=report_version,
)
def main(
    directory: Optional[str],
    username: str,
    password: Optional[str],
    auth_only: bool,
    cookie_directory: str,
    size: Sequence[AssetVersionSize],
    live_photo_size: LivePhotoVersionSize,
    recent: Optional[int],
    until_found: Optional[int],
    album: str,
    list_albums: bool,
    library: str,
    list_libraries: bool,
    skip_videos: bool,
    skip_live_photos: bool,
    xmp_sidecar: bool,
    force_size: bool,
    auto_delete: bool,
    only_print_filenames: bool,
    folder_structure: str,
    set_exif_datetime: bool,
    smtp_username: Optional[str],
    smtp_password: Optional[str],
    smtp_host: str,
    smtp_port: int,
    smtp_no_tls: bool,
    notification_email: Optional[str],
    notification_email_from: Optional[str],
    log_level: str,
    no_progress_bar: bool,
    notification_script: Optional[str],
    threads_num: int,
    delete_after_download: bool,
    keep_icloud_recent_days: Optional[int],
    domain: str,
    watch_with_interval: Optional[int],
    dry_run: bool,
    filename_cleaner: Callable[[str], str],
    lp_filename_generator: Callable[[str], str],
    raw_policy: RawTreatmentPolicy,
    password_providers: Dict[
        str, Tuple[Callable[[str], Optional[str]], Callable[[str, str], None]]
    ],
    file_match_policy: FileMatchPolicy,
    mfa_provider: MFAProvider,
    use_os_locale: bool,
) -> NoReturn:
    """Download all iCloud photos to a local directory"""

    logging.basicConfig(
        format="%(asctime)s %(levelname)-8s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        stream=sys.stdout,
    )
    logger = logging.getLogger("icloudpd")
    if only_print_filenames:
        logger.disabled = True
    else:
        # Need to make sure disabled is reset to the correct value,
        # because the logger instance is shared between tests.
        logger.disabled = False
        if log_level == "debug":
            logger.setLevel(logging.DEBUG)
        elif log_level == "info":
            logger.setLevel(logging.INFO)
        elif log_level == "error":
            logger.setLevel(logging.ERROR)

    with logging_redirect_tqdm():
        # check required directory param only if not list albums
        if not list_albums and not list_libraries and not directory and not auth_only:
            print("--auth-only, --directory, --list-libraries or --list-albums are required")
            sys.exit(2)

        if auto_delete and delete_after_download:
            print("--auto-delete and --delete-after-download are mutually exclusive")
            sys.exit(2)

        if keep_icloud_recent_days and delete_after_download:
            print(
                "--keep-icloud-recent-days and --delete-after-download should not be used together."
            )
            sys.exit(2)

        if watch_with_interval and (list_albums or only_print_filenames):  # pragma: no cover
            print(
                "--watch_with_interval is not compatible with --list_albums, --only_print_filenames"
            )
            sys.exit(2)

        # hacky way to use one param in another
        if password and "parameter" in password_providers:
            # replace
            password_providers["parameter"] = (constant(password), lambda _r, _w: None)

        if len(password_providers) == 0:  # pragma: no cover
            print("You need to specify at least one --password-provider")
            sys.exit(2)

        if "console" in password_providers and "webui" in password_providers:
            print("Console and webui are not compatible in --password-provider")
            sys.exit(2)

        if "console" in password_providers and list(password_providers)[-1] != "console":
            print("Console must be the last --password-provider")
            sys.exit(2)

        if "webui" in password_providers and list(password_providers)[-1] != "webui":
            print("Webui must be the last --password-provider")
            sys.exit(2)

        if folder_structure != "none":
            try:
                folder_structure.format(datetime.datetime.now())
            except:  # noqa E722
                print("Format specified in --folder-structure is incorrect")
                sys.exit(2)

        status_exchange = StatusExchange()
        config = Config(
            directory=directory,
            username=username,
            auth_only=auth_only,
            cookie_directory=cookie_directory,
            primary_sizes=size,
            live_photo_size=live_photo_size,
            recent=recent,
            until_found=until_found,
            album=album,
            list_albums=list_albums,
            library=library,
            list_libraries=list_libraries,
            skip_videos=skip_videos,
            skip_live_photos=skip_live_photos,
            xmp_sidecar=xmp_sidecar,
            force_size=force_size,
            auto_delete=auto_delete,
            only_print_filenames=only_print_filenames,
            folder_structure=folder_structure,
            set_exif_datetime=set_exif_datetime,
            smtp_username=smtp_username,
            smtp_host=smtp_host,
            smtp_port=smtp_port,
            smtp_no_tls=smtp_no_tls,
            notification_email=notification_email,
            notification_email_from=notification_email_from,
            log_level=log_level,
            no_progress_bar=no_progress_bar,
            notification_script=notification_script,
            threads_num=threads_num,
            delete_after_download=delete_after_download,
            keep_icloud_recent_days=keep_icloud_recent_days,
            domain=domain,
            watch_with_interval=watch_with_interval,
            dry_run=dry_run,
            raw_policy=raw_policy,
            password_providers=password_providers,
            file_match_policy=file_match_policy,
            mfa_provider=mfa_provider,
            use_os_locale=use_os_locale,
        )
        status_exchange.set_config(config)

        # hacky way to use one param in another
        if "webui" in password_providers:
            # replace
            password_providers["webui"] = (
                get_password_from_webui(logger, status_exchange),
                update_password_status_in_webui(status_exchange),
            )

        # hacky way to inject logger
        if "keyring" in password_providers:
            # replace
            password_providers["keyring"] = (
                get_password_from_keyring,
                keyring_password_writter(logger),
            )

        # start web server
        if mfa_provider == MFAProvider.WEBUI:
            server_thread = Thread(target=serve_app, daemon=True, args=[logger, status_exchange])
            server_thread.start()

        result = core(
            download_builder(
                logger,
                skip_videos,
                folder_structure,
                directory,
                size,
                force_size,
                only_print_filenames,
                set_exif_datetime,
                skip_live_photos,
                live_photo_size,
                dry_run,
                file_match_policy,
                xmp_sidecar,
            )
            if directory is not None
            else (lambda _s: lambda _c, _p: False),
            directory,
            username,
            auth_only,
            cookie_directory,
            size,
            recent,
            until_found,
            album,
            list_albums,
            library,
            list_libraries,
            skip_videos,
            auto_delete,
            only_print_filenames,
            folder_structure,
            smtp_username,
            smtp_password,
            smtp_host,
            smtp_port,
            smtp_no_tls,
            notification_email,
            notification_email_from,
            no_progress_bar,
            notification_script,
            delete_after_download,
            keep_icloud_recent_days,
            domain,
            logger,
            watch_with_interval,
            dry_run,
            filename_cleaner,
            lp_filename_generator,
            raw_policy,
            file_match_policy,
            password_providers,
            mfa_provider,
            status_exchange,
        )
        sys.exit(result)


def download_builder(
    logger: logging.Logger,
    skip_videos: bool,
    folder_structure: str,
    directory: str,
    primary_sizes: Sequence[AssetVersionSize],
    force_size: bool,
    only_print_filenames: bool,
    set_exif_datetime: bool,
    skip_live_photos: bool,
    live_photo_size: LivePhotoVersionSize,
    dry_run: bool,
    file_match_policy: FileMatchPolicy,
    xmp_sidecar: bool,
) -> Callable[[PyiCloudService], Callable[[Counter, PhotoAsset], bool]]:
    """factory for downloader"""

    def state_(icloud: PyiCloudService) -> Callable[[Counter, PhotoAsset], bool]:
        def download_photo_(counter: Counter, photo: PhotoAsset) -> bool:
            """internal function for actually downloading the photos"""

            if skip_videos and photo.item_type == AssetItemType.MOVIE:
                logger.debug(
                    "Skipping %s, only downloading photos." + "(Item type was: %s)",
                    photo.filename,
                    photo.item_type,
                )
                return False
            # Throwing error now
            # if not photo.item_type:
            #     logger.debug(
            #         "Skipping %s, only downloading photos and videos. " + "(Item type was: %s)",
            #         photo.filename,
            #         photo.item_type,
            #     )
            #     return False
            try:
                created_date = photo.created.astimezone(get_localzone())
            except (ValueError, OSError):
                logger.error(
                    "Could not convert photo created date to local timezone (%s)", photo.created
                )
                created_date = photo.created

            if folder_structure.lower() == "none":
                date_path = ""
            else:
                try:
                    date_path = folder_structure.format(created_date)
                except ValueError:  # pragma: no cover
                    # This error only seems to happen in Python 2
                    logger.error("Photo created date was not valid (%s)", photo.created)
                    # e.g. ValueError: year=5 is before 1900
                    # (https://github.com/icloud-photos-downloader/icloud_photos_downloader/issues/122)
                    # Just use the Unix epoch
                    created_date = datetime.datetime.fromtimestamp(0)
                    date_path = folder_structure.format(created_date)

            try:
                versions = disambiguate_filenames(photo.versions, primary_sizes)
            except KeyError as ex:
                print(f"KeyError: {ex} attribute was not found in the photo fields.")
                with open(file="icloudpd-photo-error.json", mode="w", encoding="utf8") as outfile:
                    json.dump(
                        {
                            "master_record": photo._master_record,
                            "asset_record": photo._asset_record,
                        },
                        outfile,
                    )
                print("icloudpd has saved the photo record to: " "./icloudpd-photo-error.json")
                print(
                    "Please create a Gist with the contents of this file: "
                    "https://gist.github.com"
                )
                print(
                    "Then create an issue on GitHub: "
                    "https://github.com/icloud-photos-downloader/icloud_photos_downloader/issues"
                )
                print(
                    "Include a link to the Gist in your issue, so that we can "
                    "see what went wrong.\n"
                )
                return False

            download_dir = os.path.normpath(os.path.join(directory, date_path))
            success = False

            for download_size in primary_sizes:
                if download_size not in versions and download_size != AssetVersionSize.ORIGINAL:
                    if force_size:
                        logger.error(
                            "%s size does not exist for %s. Skipping...",
                            download_size.value,
                            photo.filename,
                        )
                        continue
                    if AssetVersionSize.ORIGINAL in primary_sizes:
                        continue  # that should avoid double download for original
                    download_size = AssetVersionSize.ORIGINAL

                version = versions[download_size]
                filename = version.filename

                download_path = local_download_path(filename, download_dir)

                original_download_path = None
                file_exists = os.path.isfile(download_path)
                if not file_exists and download_size == AssetVersionSize.ORIGINAL:
                    # Deprecation - We used to download files like IMG_1234-original.jpg,
                    # so we need to check for these.
                    # Now we match the behavior of iCloud for Windows: IMG_1234.jpg
                    original_download_path = add_suffix_to_filename("-original", download_path)
                    file_exists = os.path.isfile(original_download_path)

                if file_exists:
                    if file_match_policy == FileMatchPolicy.NAME_SIZE_DEDUP_WITH_SUFFIX:
                        # for later: this crashes if download-size medium is specified
                        file_size = os.stat(original_download_path or download_path).st_size
                        photo_size = version.size
                        if file_size != photo_size:
                            download_path = (f"-{photo_size}.").join(download_path.rsplit(".", 1))
                            logger.debug("%s deduplicated", truncate_middle(download_path, 96))
                            file_exists = os.path.isfile(download_path)
                    if file_exists:
                        counter.increment()
                        logger.debug("%s already exists", truncate_middle(download_path, 96))

                if not file_exists:
                    counter.reset()
                    if only_print_filenames:
                        print(download_path)
                    else:
                        truncated_path = truncate_middle(download_path, 96)
                        logger.debug("Downloading %s...", truncated_path)

                        download_result = download.download_media(
                            logger, dry_run, icloud, photo, download_path, version, download_size
                        )
                        success = download_result

                        if download_result:
                            if (
                                not dry_run
                                and set_exif_datetime
                                and filename.lower().endswith((".jpg", ".jpeg"))
                                and not exif_datetime.get_photo_exif(logger, download_path)
                            ):
                                # %Y:%m:%d looks wrong, but it's the correct format
                                date_str = created_date.strftime("%Y-%m-%d %H:%M:%S%z")
                                logger.debug(
                                    "Setting EXIF timestamp for %s: %s", download_path, date_str
                                )
                                exif_datetime.set_photo_exif(
                                    logger,
                                    download_path,
                                    created_date.strftime("%Y:%m:%d %H:%M:%S"),
                                )
                            if not dry_run:
                                download.set_utime(download_path, created_date)
                            logger.info("Downloaded %s", truncated_path)

                if xmp_sidecar:
                    generate_xmp_file(logger, download_path, photo._asset_record)

            # Also download the live photo if present
            if not skip_live_photos:
                lp_size = live_photo_size
                if lp_size in photo.versions:
                    version = photo.versions[lp_size]
                    lp_filename = version.filename
                    if live_photo_size != LivePhotoVersionSize.ORIGINAL:
                        # Add size to filename if not original
                        lp_filename = add_suffix_to_filename(
                            size_to_suffix(live_photo_size),
                            lp_filename,
                        )
                    else:
                        pass
                    lp_download_path = os.path.join(download_dir, lp_filename)

                    lp_file_exists = os.path.isfile(lp_download_path)

                    if only_print_filenames and not lp_file_exists:
                        print(lp_download_path)
                    else:
                        if lp_file_exists:
                            if file_match_policy == FileMatchPolicy.NAME_SIZE_DEDUP_WITH_SUFFIX:
                                lp_file_size = os.stat(lp_download_path).st_size
                                lp_photo_size = version.size
                                if lp_file_size != lp_photo_size:
                                    lp_download_path = (f"-{lp_photo_size}.").join(
                                        lp_download_path.rsplit(".", 1)
                                    )
                                    logger.debug(
                                        "%s deduplicated", truncate_middle(lp_download_path, 96)
                                    )
                                    lp_file_exists = os.path.isfile(lp_download_path)
                            if lp_file_exists:
                                logger.debug(
                                    "%s already exists", truncate_middle(lp_download_path, 96)
                                )
                        if not lp_file_exists:
                            truncated_path = truncate_middle(lp_download_path, 96)
                            logger.debug("Downloading %s...", truncated_path)
                            download_result = download.download_media(
                                logger, dry_run, icloud, photo, lp_download_path, version, lp_size
                            )
                            success = download_result and success
                            if download_result:
                                logger.info("Downloaded %s", truncated_path)
            return success

        return download_photo_

    return state_


def delete_photo(
    logger: logging.Logger,
    photo_service: PhotosService,
    library_object: PhotoLibrary,
    photo: PhotoAsset,
) -> None:
    """Delete a photo from the iCloud account."""
    clean_filename_local = photo.filename
    logger.debug("Deleting %s in iCloud...", clean_filename_local)
    url = (
        f"{photo_service._service_endpoint}/records/modify?"
        f"{urllib.parse.urlencode(photo_service.params)}"
    )
    post_data = json.dumps(
        {
            "atomic": True,
            "desiredKeys": ["isDeleted"],
            "operations": [
                {
                    "operationType": "update",
                    "record": {
                        "fields": {"isDeleted": {"value": 1}},
                        "recordChangeTag": photo._asset_record["recordChangeTag"],
                        "recordName": photo._asset_record["recordName"],
                        "recordType": "CPLAsset",
                    },
                }
            ],
            "zoneID": library_object.zone_id,
        }
    )
    photo_service.session.post(url, data=post_data, headers={"Content-type": "application/json"})
    logger.info("Deleted %s in iCloud", clean_filename_local)


def delete_photo_dry_run(
    logger: logging.Logger,
    _photo_service: PhotosService,
    library_object: PhotoLibrary,
    photo: PhotoAsset,
) -> None:
    """Dry run for deleting a photo from the iCloud"""
    logger.info(
        "[DRY RUN] Would delete %s in iCloud library %s",
        photo.filename,
        library_object.zone_id["zoneName"],
    )


RetrierT = TypeVar("RetrierT")


def retrier(
    func: Callable[[], RetrierT], error_handler: Callable[[Exception, int], None]
) -> RetrierT:
    """Run main func and retry helper if receive session error"""
    attempts = 0
    while True:
        try:
            return func()
        except Exception as ex:
            attempts += 1
            error_handler(ex, attempts)
            if attempts > constants.MAX_RETRIES:
                raise


def session_error_handle_builder(
    logger: Logger, icloud: PyiCloudService
) -> Callable[[Exception, int], None]:
    """Build handler for session error"""

    def session_error_handler(ex: Exception, attempt: int) -> None:
        """Handles session errors in the PhotoAlbum photos iterator"""
        if "Invalid global session" in str(ex):
            if attempt > constants.MAX_RETRIES:
                logger.error("iCloud re-authentication failed. Please try again later.")
                raise ex
            logger.error("Session error, re-authenticating...")
            if attempt > 1:
                # If the first re-authentication attempt failed,
                # start waiting a few seconds before retrying in case
                # there are some issues with the Apple servers
                time.sleep(constants.WAIT_SECONDS * attempt)
            icloud.authenticate()

    return session_error_handler


def internal_error_handle_builder(logger: logging.Logger) -> Callable[[Exception, int], None]:
    """Build handler for internal error"""

    def internal_error_handler(ex: Exception, attempt: int) -> None:
        """Handles session errors in the PhotoAlbum photos iterator"""
        if "INTERNAL_ERROR" in str(ex):
            if attempt > constants.MAX_RETRIES:
                logger.error("Internal Error at Apple.")
                raise ex
            logger.error("Internal Error at Apple, retrying...")
            # start waiting a few seconds before retrying in case
            # there are some issues with the Apple servers
            time.sleep(constants.WAIT_SECONDS * attempt)

    return internal_error_handler


def compose_handlers(
    handlers: Sequence[Callable[[Exception, int], None]],
) -> Callable[[Exception, int], None]:
    """Compose multiple error handlers"""

    def composed(ex: Exception, retries: int) -> None:
        for handler in handlers:
            handler(ex, retries)

    return composed


def core(
    downloader: Callable[[PyiCloudService], Callable[[Counter, PhotoAsset], bool]],
    directory: Optional[str],
    username: str,
    auth_only: bool,
    cookie_directory: str,
    primary_sizes: Sequence[AssetVersionSize],
    recent: Optional[int],
    until_found: Optional[int],
    album: str,
    list_albums: bool,
    library: str,
    list_libraries: bool,
    skip_videos: bool,
    auto_delete: bool,
    only_print_filenames: bool,
    folder_structure: str,
    smtp_username: Optional[str],
    smtp_password: Optional[str],
    smtp_host: str,
    smtp_port: int,
    smtp_no_tls: bool,
    notification_email: Optional[str],
    notification_email_from: Optional[str],
    no_progress_bar: bool,
    notification_script: Optional[str],
    delete_after_download: bool,
    keep_icloud_recent_days: Optional[int],
    domain: str,
    logger: logging.Logger,
    watch_interval: Optional[int],
    dry_run: bool,
    filename_cleaner: Callable[[str], str],
    lp_filename_generator: Callable[[str], str],
    raw_policy: RawTreatmentPolicy,
    file_match_policy: FileMatchPolicy,
    password_providers: Dict[
        str, Tuple[Callable[[str], Optional[str]], Callable[[str, str], None]]
    ],
    mfa_provider: MFAProvider,
    status_exchange: StatusExchange,
) -> int:
    """Download all iCloud photos to a local directory"""

    raise_error_on_2sa = (
        smtp_username is not None
        or notification_email is not None
        or notification_script is not None
    )
    try:
        icloud = authenticator(
            logger,
            domain,
            filename_cleaner,
            lp_filename_generator,
            raw_policy,
            file_match_policy,
            password_providers,
            mfa_provider,
            status_exchange,
        )(
            username,
            cookie_directory,
            raise_error_on_2sa,
            os.environ.get("CLIENT_ID"),
        )
    except TwoStepAuthRequiredError:
        if notification_script is not None:
            subprocess.call([notification_script])
        if smtp_username is not None or notification_email is not None:
            send_2sa_notification(
                logger,
                smtp_username,
                smtp_password,
                smtp_host,
                smtp_port,
                smtp_no_tls,
                notification_email,
                notification_email_from,
            )
        return 1

    if auth_only:
        logger.info("Authentication completed successfully")
        return 0

    download_photo = downloader(icloud)

    # Access to the selected library. Defaults to the primary photos object.
    library_object: PhotoLibrary = icloud.photos

    if list_libraries:
        libraries_dict = icloud.photos.libraries
        library_names = libraries_dict.keys()
        print(*library_names, sep="\n")

    else:
        while True:
            # Default album is "All Photos", so this is the same as
            # calling `icloud.photos.all`.
            # After 6 or 7 runs within 1h Apple blocks the API for some time. In that
            # case exit.
            try:
                if library:
                    try:
                        library_object = icloud.photos.libraries[library]
                    except KeyError:
                        logger.error("Unknown library: %s", library)
                        return 1
                photos = library_object.albums[album]
            except PyiCloudAPIResponseException as err:
                # For later: come up with a nicer message to the user. For now take the
                # exception text
                logger.error("error?? %s", err)
                return 1

            if list_albums:
                print("Albums:")
                albums_dict = library_object.albums
                albums = albums_dict.values()  # pragma: no cover
                album_titles = [str(a) for a in albums]
                print(*album_titles, sep="\n")
                return 0
            # casting is okay since we checked for list_albums and directory compatibily upstream
            # would be better to have that in types though
            directory = os.path.normpath(cast(str, directory))

            videos_phrase = "" if skip_videos else " and videos"
            logger.debug("Looking up all photos%s from album %s...", videos_phrase, album)

            session_exception_handler = session_error_handle_builder(logger, icloud)
            internal_error_handler = internal_error_handle_builder(logger)

            error_handler = compose_handlers([session_exception_handler, internal_error_handler])

            photos.exception_handler = error_handler

            photos_count: Optional[int] = len(photos)

            photos_enumerator: Iterable[PhotoAsset] = photos

            # Optional: Only download the x most recent photos.
            if recent is not None:
                photos_count = recent
                photos_enumerator = itertools.islice(photos_enumerator, recent)

            if until_found is not None:
                photos_count = None
                # ensure photos iterator doesn't have a known length
                photos_enumerator = (p for p in photos_enumerator)

            # Skip the one-line progress bar if we're only printing the filenames,
            # or if the progress bar is explicitly disabled,
            # or if this is not a terminal (e.g. cron or piping output to file)
            skip_bar = not os.environ.get("FORCE_TQDM") and (
                only_print_filenames or no_progress_bar or not sys.stdout.isatty()
            )
            if skip_bar:
                photos_enumerator = photos_enumerator
                # logger.set_tqdm(None)
            else:
                photos_enumerator = tqdm(
                    iterable=photos_enumerator,
                    total=photos_count,
                    leave=False,
                    dynamic_ncols=True,
                    ascii=True,
                )
                # logger.set_tqdm(photos_enumerator)

            if photos_count is not None:
                plural_suffix = "" if photos_count == 1 else "s"
                video_suffix = ""
                photos_count_str = "the first" if photos_count == 1 else photos_count

                if not skip_videos:
                    video_suffix = " or video" if photos_count == 1 else " and videos"
            else:
                photos_count_str = "???"
                plural_suffix = "s"
                video_suffix = " and videos" if not skip_videos else ""
            logger.info(
                ("Downloading %s %s" + " photo%s%s to %s ..."),
                photos_count_str,
                ",".join([_s.value for _s in primary_sizes]),
                plural_suffix,
                video_suffix,
                directory,
            )

            consecutive_files_found = Counter(0)

            def should_break(counter: Counter) -> bool:
                """Exit if until_found condition is reached"""
                return until_found is not None and counter.value() >= until_found

            status_exchange.get_progress().photos_count = (
                0 if photos_count is None else photos_count
            )
            photos_counter = 0

            photos_iterator = iter(photos_enumerator)
            while True:
                try:
                    if should_break(consecutive_files_found):
                        logger.info(
                            "Found %s consecutive previously downloaded photos. Exiting",
                            until_found,
                        )
                        break
                    item = next(photos_iterator)
                    if download_photo(consecutive_files_found, item) and delete_after_download:
                        delete_local = partial(
                            delete_photo_dry_run if dry_run else delete_photo,
                            logger,
                            icloud.photos,
                            library_object,
                            item,
                        )

                        retrier(delete_local, error_handler)

                    photos_counter += 1
                    status_exchange.get_progress().photos_counter = photos_counter

                    if status_exchange.get_progress().cancel:
                        break

                except StopIteration:
                    break

            if only_print_filenames:
                return 0

            if status_exchange.get_progress().cancel:
                logger.info("Iteration was cancelled")
                status_exchange.get_progress().photos_last_message = "Iteration was cancelled"
            else:
                logger.info("All photos have been downloaded")
                status_exchange.get_progress().photos_last_message = (
                    "All photos have been downloaded"
                )
            status_exchange.get_progress().reset()

            if auto_delete:
                autodelete_photos(
                    logger, dry_run, library_object, folder_structure, directory, primary_sizes
                )

            if keep_icloud_recent_days is not None:
                try:
                    now = datetime.datetime.now(get_localzone())
                    created_date = item.created.astimezone(get_localzone())
                    age_days = (now - created_date).days
                    if age_days < keep_icloud_recent_days:
                        logger.debug(
                            "Skipping deletion of %s as it is within the keep_icloud_recent_days period (%d days old)",
                            item.filename,
                            age_days,
                        )
                    else:
                        delete_local = partial(
                            delete_photo_dry_run if dry_run else delete_photo,
                            logger,
                            icloud.photos,
                            library_object,
                            item,
                        )

                        retrier(delete_local, error_handler)
                        logger.debug(
                            "Deleted %s as it is older than the keep_icloud_recent_days period (%d days old)",
                            item.filename,
                            age_days,
                        )
                except (ValueError, OSError):
                    logger.error(
                        "Could not convert photo created date to local timezone (%s)",
                        item.created,
                    )
                except Exception as e:
                    logger.error(f"Error deleting photo: {e}")

            if watch_interval:  # pragma: no cover
                logger.info(f"Waiting for {watch_interval} sec...")
                interval: Sequence[int] = range(1, watch_interval)
                iterable: Sequence[int] = (
                    interval
                    if skip_bar
                    else typing.cast(
                        Sequence[int],
                        tqdm(
                            iterable=interval,
                            desc="Waiting...",
                            ascii=True,
                            leave=False,
                            dynamic_ncols=True,
                        ),
                    )
                )
                for counter in iterable:
                    status_exchange.get_progress().waiting = watch_interval - counter
                    if status_exchange.get_progress().resume:
                        status_exchange.get_progress().reset()
                        break
                    time.sleep(1)
            else:
                break  # pragma: no cover

    return 0



import bcrypt
import os


def authenticate_secret(password: str, path: str) -> bool:
    if not os.path.exists(path):
        if password == "":
            return True
        else:
            raise FileNotFoundError("Secret file missing, please try resetting the server.")
    with open(path, "rb") as f:
        secret_hash = f.read()
    return bcrypt.checkpw(password.encode(), secret_hash)


def save_secret_hash(password: str, path: str) -> None:
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "wb") as f:
        f.write(bcrypt.hashpw(password.encode(), bcrypt.gensalt()))
    # Set file permissions to read/write for owner only (600)
    os.chmod(path, 0o600)



from html.parser import HTMLParser
import json
from typing import Any, List, Mapping, NamedTuple, Optional, Protocol, Sequence, Tuple

class _SMSParser(HTMLParser):
    def __init__(self) -> None:
        # initialize the base class
        super(_SMSParser, self).__init__()
        self._is_boot_args = False
        self.sms_data: Mapping[str, Any] = {}

    def handle_starttag(self, tag: str, attrs: List[Tuple[str, Optional[str]]]) -> None:
        if tag == "script":
            self._is_boot_args = ("type", "application/json") in attrs and ("class", "boot_args") in attrs

    def handle_endtag(self, tag: str) -> None:
        if tag == "script":
            self._is_boot_args = False

    def handle_data(self, data: str) -> None:
        if self._is_boot_args:
            self.sms_data = json.loads(data)

class TrustedDevice(Protocol):
    @property
    def id(self) -> int: ...
    @property
    def obfuscated_number(self) -> str: ...

class _InternalTrustedDevice(NamedTuple):
    id: int
    obfuscated_number: str

def _map_to_trusted_device(device: Mapping[str, Any]) -> Optional[TrustedDevice]:
    id: Optional[int] = device.get("id")
    number: Optional[str] = device.get("obfuscatedNumber")
    if id is None or number is None:
        return None
    return _InternalTrustedDevice(id=id, obfuscated_number=number.replace('•', '*'))

class _Response(Protocol):
    @property
    def status_code(self) -> int: ...
    @property
    def text(self) -> str: ...

def parse_trusted_phone_numbers_response(response: _Response) -> Sequence[TrustedDevice]:
    """ Parses html response for the list of available trusted phone numbers"""
    if response.status_code in [200, 204]:
        return parse_trusted_phone_numbers_payload(response.text)
    return []

def parse_trusted_phone_numbers_payload(content: str) -> Sequence[TrustedDevice]:
    """ Parses html response for the list of available trusted phone numbers"""
    parser = _SMSParser()
    parser.feed(content)
    parser.close()
    numbers: Sequence[Mapping[str, Any]] = parser.sms_data.get("direct", {}).get("twoSV", {}).get("phoneNumberVerification", {}).get("trustedPhoneNumbers", [])
    return list((item for item in map(_map_to_trusted_device, numbers) if item is not None))

class AuthenticatedSession(NamedTuple):
    client_id: str
    scnt: str
    session_id: str

def _oauth_const_headers() -> Mapping[str, str]:
    return {
        "X-Apple-OAuth-Client-Id": "d39ba9916b7251055b22c7f910e2ea796ee65e98b2ddecea8f5dde8d9d1a815d",
        "X-Apple-OAuth-Client-Type": "firstPartyAuth",
        "X-Apple-OAuth-Require-Grant-Code": "true",
        "X-Apple-Widget-Key": "d39ba9916b7251055b22c7f910e2ea796ee65e98b2ddecea8f5dde8d9d1a815d",
    }

def _oauth_redirect_header(domain: str) -> Mapping[str, str]:
    return {
        "X-Apple-OAuth-Redirect-URI": "https://www.icloud.com.cn" if domain == "cn" else "https://www.icloud.com",
    }

def _oauth_headers(auth_session: AuthenticatedSession) -> Mapping[str, str]:
    """ Headers with OAuth session """

    return {
        "X-Apple-OAuth-State": auth_session.client_id,
        "scnt": auth_session.scnt,
        "X-Apple-ID-Session-Id": auth_session.session_id
    }

def _auth_url(domain: str) -> str:
    return "https://idmsa.apple.com.cn/appleauth/auth" if domain == "cn" else "https://idmsa.apple.com/appleauth/auth"

class _DomainProvider(Protocol):
    @property
    def domain(self) -> str: ...

class _OAuthSessionProvider(Protocol):
    @property
    def oauth_session(self) -> AuthenticatedSession: ...

class _TrustedPhoneContextProvider(_DomainProvider, _OAuthSessionProvider, Protocol): ...

class Request(Protocol):
    @property
    def method(self) -> str: ...
    @property
    def url(self) -> str: ...
    @property
    def headers(self) -> Mapping[str, str]: ...
    @property
    def data(self) -> Optional[str]: ...
    @property
    def json(self) -> Optional[Mapping[str, Any]]: ...

class _InternalRequest(NamedTuple):
    method: str
    url: str
    headers: Mapping[str, str]
    data: Optional[str] = None
    json: Optional[Mapping[str, Any]] = None

def build_trusted_phone_numbers_request(context: _TrustedPhoneContextProvider) -> Request:
    """ Builds a request for the list of trusted phone numbers for sms 2fa """

    url = _auth_url(context.domain)

    req = _InternalRequest(
        method="GET",
        url=url,
        headers = {
            **_oauth_const_headers(),
            **_oauth_redirect_header(context.domain),
            **_oauth_headers(context.oauth_session),
        })
    return req

def build_send_sms_code_request(context: _TrustedPhoneContextProvider, device_id: int) -> Request:
    """ Builds a request for the list of trusted phone numbers for sms 2fa """

    url = _auth_url(context.domain) + "/verify/phone"

    json = {"phoneNumber":{"id":device_id},"mode":"sms"}

    req = _InternalRequest(
        method="PUT",
        url=url,
        headers = {
            **_oauth_const_headers(),
            **_oauth_redirect_header(context.domain),
            **_oauth_headers(context.oauth_session),
            ** {"Content-type": "application/json; charset=utf-8"},
        },
        json = json)
    return req

def build_verify_sms_code_request(context: _TrustedPhoneContextProvider, device_id: int, code: str) -> Request:
    """ Builds a request for the list of trusted phone numbers for sms 2fa """

    url = _auth_url(context.domain) + "/verify/phone/securitycode"

    json = {"phoneNumber":{"id":device_id},"securityCode":{"code":code},"mode":"sms"}

    req = _InternalRequest(
        method="POST",
        url=url,
        headers = {
            **_oauth_const_headers(),
            **_oauth_redirect_header(context.domain),
            **_oauth_headers(context.oauth_session),
            ** {"Content-type": "application/json; charset=utf-8"},
            ** {"Accept": "application/json; charset=utf-8"},
        },
        json = json)
    return req



from pydantic import BaseModel, Field
from typing import Sequence, Literal, Annotated
from enum import Enum

NON_POLICY_FIELDS = ["status", "progress", "authenticated", "albums"]


class AuthenticationResult(Enum):
    SUCCESS = "success"
    FAILED = "failed"
    MFA_REQUIRED = "mfa_required"


class PolicyConfigs(BaseModel):
    # Connection options
    username: str = Field(min_length=1)
    domain: Literal["com", "cn"] = "com"

    # Download options
    directory: str = Field(min_length=1)
    folder_structure: str = "{:%Y/%m/%d}"
    size: Sequence[Literal["original", "medium", "thumb", "adjusted", "alternative"]] = ["original"]
    live_photo_size: Literal["original", "medium", "thumb"] = "original"
    force_size: bool = False
    align_raw: Literal["original", "alternative", "as-is"] = "original"
    keep_unicode_in_filenames: bool = False
    set_exif_datetime: bool = False
    live_photo_mov_filename_policy: Literal["original", "suffix"] = "suffix"
    file_match_policy: Literal["name-size-dedup-with-suffix", "name-id7"] = (
        "name-size-dedup-with-suffix"
    )
    xmp_sidecar: bool = False
    use_os_locale: bool = False

    # Filter options
    album: str = "All Photos"
    library: Literal["Personal Library", "Shared Library"] = "Personal Library"
    recent: Annotated[int, Field(ge=0)] | None = None
    until_found: Annotated[int, Field(ge=0)] | None = None
    skip_videos: bool = False
    skip_live_photos: bool = False

    # Delete options
    auto_delete: bool = False
    delete_after_download: bool = False

    # icloudpd-ui options
    interval: Annotated[int, Field(ge=0)] | None = None
    log_level: Literal["debug", "info", "error"] = "info"



import datetime


class Progress:
    def __init__(self) -> None:
        self._photos_count = 0
        self._photos_counter = 0
        self.photos_percent = 0
        self.photos_last_message = ""
        self.waiting_readable = ""
        self.resume = False
        self.cancel = False
        self._waiting = 0

    @property
    def waiting(self) -> int:
        return self._waiting

    @waiting.setter
    def waiting(self, waiting: int) -> None:
        self._waiting = waiting
        self.waiting_readable = str(datetime.timedelta(seconds=waiting))

    @property
    def photos_count(self) -> int:
        return self._photos_count

    @photos_count.setter
    def photos_count(self, photos_count: int) -> None:
        self._photos_count = photos_count
        if self.photos_count != 0:
            self.photos_percent = round(100 / self.photos_count * self.photos_counter)
        else:
            self.photos_percent = 0

    @property
    def photos_counter(self) -> int:
        return self._photos_counter

    @photos_counter.setter
    def photos_counter(self, photos_counter: int) -> None:
        self._photos_counter = photos_counter
        if self.photos_count != 0:
            self.photos_percent = round(100 / self.photos_count * self.photos_counter)
        else:
            self.photos_percent = 0

    def reset(self) -> None:
        self._photos_count = 0
        self._photos_counter = 0
        self.photos_percent = 0
        self._waiting = 0
        self.waiting_readable = ""
        self.resume = False
        self.cancel = False



"""String helper functions"""


def truncate_middle(string: str, length: int) -> str:
    """Truncates a string to a maximum length, inserting "..." in the middle"""
    if len(string) <= length:
        return string
    if length < 0:
        raise ValueError("n must be greater than or equal to 1")
    if length <= 3:
        return "..."[0:length]
    end_length = int(length) // 2 - 2
    start_length = length - end_length - 4
    end_length = max(end_length, 1)
    return f"{string[:start_length]}...{string[-end_length:]}"



import copy
import os
from typing import Dict, Optional, Sequence
import typing
import keyring

from pyicloud_ipd.asset_version import AssetVersion
from pyicloud_ipd.version_size import AssetVersionSize, VersionSize

from .exceptions import PyiCloudNoStoredPasswordAvailableException

KEYRING_SYSTEM = 'pyicloud://icloud-password'


# def get_password(username:str, interactive:bool=sys.stdout.isatty()) -> str:
#     try:
#         return get_password_from_keyring(username)
#     except PyiCloudNoStoredPasswordAvailableException:
#         if not interactive:
#             raise

#         return getpass.getpass(
#             'Enter iCloud password for {username}: '.format(
#                 username=username,
#             )
#         )


def password_exists_in_keyring(username:str) -> bool:
    try:
        return get_password_from_keyring(username) is not None
    except PyiCloudNoStoredPasswordAvailableException:
        return False


def get_password_from_keyring(username:str) -> Optional[str]:
    result = keyring.get_password(
        KEYRING_SYSTEM,
        username
    )
    # if result is None:
    #     raise PyiCloudNoStoredPasswordAvailableException(
    #         "No pyicloud password for {username} could be found "
    #         "in the system keychain.  Use the `--store-in-keyring` "
    #         "command-line option for storing a password for this "
    #         "username.".format(
    #             username=username,
    #         )
    #     )

    return result


def store_password_in_keyring(username: str, password:str) -> None:
    # if get_password_from_keyring(username) is not None:
    #     # Apple can save only into empty keyring
    #     return delete_password_in_keyring(username)
    return keyring.set_password(
        KEYRING_SYSTEM,
        username,
        password,
    )


def delete_password_in_keyring(username:str) -> None:
    return keyring.delete_password(
        KEYRING_SYSTEM,
        username,
    )


def underscore_to_camelcase(word:str , initial_capital: bool=False) -> str:
    words = [x.capitalize() or '_' for x in word.split('_')]
    if not initial_capital:
        words[0] = words[0].lower()

    return ''.join(words)


# def filename_with_size(filename: str, size: str, original: Optional[Dict[str, Any]]) -> str:
#     """Returns the filename with size, e.g. IMG1234.jpg, IMG1234-small.jpg"""
#     if size == 'original' or size == 'alternative':
#         # TODO what if alternative ext matches original?
#         return filename
#     # for adjustments we add size only if extension matches original (alternative
#     if size == "adjusted" and original != None and original["filename"] != filename:
#         return filename
#     return (f"-{size}.").join(filename.rsplit(".", 1))

def size_to_suffix(size: VersionSize) -> str:
    return f"-{size}".lower()

def add_suffix_to_filename(suffix: str, filename: str) -> str:
    _n, _e = os.path.splitext(filename)
    return _n + suffix + _e

def disambiguate_filenames(_versions: Dict[VersionSize, AssetVersion], _sizes:Sequence[AssetVersionSize]) -> Dict[AssetVersionSize, AssetVersion]:
    _results: Dict[AssetVersionSize, AssetVersion] = {}
    # add those that were requested
    for _size in _sizes:
        _version = _versions.get(_size)
        if _version:
            _results[_size] = copy.copy(_version)

    # adjusted
    if AssetVersionSize.ADJUSTED in _sizes:
        if AssetVersionSize.ORIGINAL not in _sizes:
            if AssetVersionSize.ADJUSTED not in _results:
                # clone
                _results[AssetVersionSize.ADJUSTED] = copy.copy(_versions[AssetVersionSize.ORIGINAL])
        else:
            if AssetVersionSize.ADJUSTED in _results and _results[AssetVersionSize.ORIGINAL].filename == _results[AssetVersionSize.ADJUSTED].filename:
                _results[AssetVersionSize.ADJUSTED].filename = add_suffix_to_filename("-adjusted", _results[AssetVersionSize.ADJUSTED].filename)

    # alternative
    if AssetVersionSize.ALTERNATIVE in _sizes:
        if AssetVersionSize.ORIGINAL not in _sizes and AssetVersionSize.ADJUSTED not in _results:
            if AssetVersionSize.ALTERNATIVE not in _results:
                # clone
                _results[AssetVersionSize.ALTERNATIVE] = copy.copy(_versions[AssetVersionSize.ORIGINAL])
        else:
            if AssetVersionSize.ALTERNATIVE in _results:
                if AssetVersionSize.ADJUSTED in _results and _results[AssetVersionSize.ADJUSTED].filename == _results[AssetVersionSize.ALTERNATIVE].filename or AssetVersionSize.ORIGINAL in _results and _results[AssetVersionSize.ORIGINAL].filename == _results[AssetVersionSize.ALTERNATIVE].filename:
                    _results[AssetVersionSize.ALTERNATIVE].filename = add_suffix_to_filename("-alternative", _results[AssetVersionSize.ALTERNATIVE].filename)

    for _size in _sizes:
        if _size not in [AssetVersionSize.ORIGINAL, AssetVersionSize.ADJUSTED, AssetVersionSize.ALTERNATIVE]:
            if _size not in _results:
                # ensure original is downloaded - mimic existing behavior
                if AssetVersionSize.ORIGINAL not in _sizes:
                    _results[AssetVersionSize.ORIGINAL] = copy.copy(_versions[AssetVersionSize.ORIGINAL])
            # else:
            #     _n, _e = os.path.splitext(_results[_size]["filename"])
            #     _results[_size]["filename"] = f"{_n}-{_size}{_e}"


    return _results



import sys
from typing import Any, Callable, Dict, NamedTuple, Optional, Sequence
import typing
from uuid import uuid1
import json
import logging
from tempfile import gettempdir
from os import path, mkdir
from re import match
import http.cookiejar as cookielib
import getpass
import srp
import base64
import hashlib

from requests import PreparedRequest, Request, Response

from pyicloud_ipd.exceptions import (
    PyiCloudConnectionException,
    PyiCloudFailedLoginException,
    PyiCloudAPIResponseException,
    PyiCloudServiceNotActivatedException,
)
from pyicloud_ipd.file_match import FileMatchPolicy
from pyicloud_ipd.raw_policy import RawTreatmentPolicy
from pyicloud_ipd.services.findmyiphone import AppleDevice, FindMyiPhoneServiceManager
from pyicloud_ipd.services.calendar import CalendarService
from pyicloud_ipd.services.ubiquity import UbiquityService
from pyicloud_ipd.services.contacts import ContactsService
from pyicloud_ipd.services.reminders import RemindersService
from pyicloud_ipd.services.photos import PhotosService
from pyicloud_ipd.services.account import AccountService
from pyicloud_ipd.session import PyiCloudPasswordFilter, PyiCloudSession
from pyicloud_ipd.sms import AuthenticatedSession, TrustedDevice, build_send_sms_code_request, build_trusted_phone_numbers_request, build_verify_sms_code_request, parse_trusted_phone_numbers_response


LOGGER = logging.getLogger(__name__)

HEADER_DATA = {
    "X-Apple-ID-Account-Country": "account_country",
    "X-Apple-ID-Session-Id": "session_id",
    "X-Apple-Session-Token": "session_token",
    "X-Apple-TwoSV-Trust-Token": "trust_token",
    "X-Apple-TwoSV-Trust-Eligible": "trust_eligible",
    "X-Apple-I-Rscd": "apple_rscd",
    "X-Apple-I-Ercd": "apple_ercd",
    "scnt": "scnt",
}

class TrustedPhoneContextProvider(NamedTuple):
    domain: str
    oauth_session: AuthenticatedSession



class PyiCloudService:
    """
    A base authentication class for the iCloud service. Handles the
    authentication required to access iCloud services.

    Usage:
        from pyicloud_ipd import PyiCloudService
        pyicloud = PyiCloudService('username@apple.com', 'password')
        pyicloud_ipd.iphone.location()
    """

    def __init__(
        self, 
        filename_cleaner: Callable[[str], str],
        lp_filename_generator: Callable[[str], str],
        domain:str, 
        raw_policy: RawTreatmentPolicy,
        file_match_policy: FileMatchPolicy,
        apple_id: str, password:str, cookie_directory:Optional[str]=None, verify:bool=True,
        client_id:Optional[str]=None, with_family:bool=True,
    ):
        self.filename_cleaner = filename_cleaner
        self.lp_filename_generator = lp_filename_generator
        self.raw_policy = raw_policy
        self.file_match_policy = file_match_policy
        self.user: Dict[str, Any] = {"accountName": apple_id, "password": password}
        self.data: Dict[str, Any] = {} 
        self.params: Dict[str, Any] = {}
        self.client_id: str = client_id or ("auth-%s" % str(uuid1()).lower())
        self.with_family = with_family

        self.password_filter = PyiCloudPasswordFilter(password)
        LOGGER.addFilter(self.password_filter)

        if (domain == 'com'):
            self.AUTH_ENDPOINT = "https://idmsa.apple.com/appleauth/auth"
            self.HOME_ENDPOINT = "https://www.icloud.com"
            self.SETUP_ENDPOINT = "https://setup.icloud.com/setup/ws/1"
        elif (domain == 'cn'):
            self.AUTH_ENDPOINT = "https://idmsa.apple.com.cn/appleauth/auth"
            self.HOME_ENDPOINT = "https://www.icloud.com.cn"
            self.SETUP_ENDPOINT = "https://setup.icloud.com.cn/setup/ws/1"
        else:
            raise NotImplementedError(f"Domain '{domain}' is not supported yet")

        self.domain = domain

        if cookie_directory:
            self._cookie_directory = path.expanduser(
                path.normpath(cookie_directory)
            )
            if not path.exists(self._cookie_directory):
                mkdir(self._cookie_directory, 0o700)
        else:
            topdir = path.join(gettempdir(), "pyicloud")
            self._cookie_directory = path.join(topdir, getpass.getuser())
            if not path.exists(topdir):
                mkdir(topdir, 0o777)
            if not path.exists(self._cookie_directory):
                mkdir(self._cookie_directory, 0o700)

        LOGGER.debug("Using session file %s", self.session_path)

        self.session_data = {}
        try:
            with open(self.session_path, encoding="utf-8") as session_f:
                self.session_data = json.load(session_f)
        except:  
            LOGGER.info("Session file does not exist")
        session_client_id: Optional[str] = self.session_data.get("client_id")
        if session_client_id:
            self.client_id = session_client_id
        else:
            self.session_data.update({"client_id": self.client_id})

        self.session:PyiCloudSession = PyiCloudSession(self)
        self.session.verify = verify
        self.session.headers.update({
            'Origin': self.HOME_ENDPOINT,
            'Referer': '%s/' % self.HOME_ENDPOINT,
            'User-Agent': 'Opera/9.52 (X11; Linux i686; U; en)'
        })

        cookiejar_path = self.cookiejar_path
        self.session.cookies = cookielib.LWPCookieJar(filename=cookiejar_path) # type: ignore[assignment]
        if path.exists(cookiejar_path):
            try:
                self.session.cookies.load(ignore_discard=True, ignore_expires=True) # type: ignore[attr-defined]
                LOGGER.debug("Read cookies from %s", cookiejar_path)
            except:
                # Most likely a pickled cookiejar from earlier versions.
                # The cookiejar will get replaced with a valid one after
                # successful authentication.
                LOGGER.warning("Failed to read cookiejar %s", cookiejar_path)

        # Unsure if this is still needed
        self.params = {
            'clientBuildNumber': '17DHotfix5',
            'clientMasteringNumber': '17DHotfix5',
            'ckjsBuildVersion': '17DProjectDev77',
            'ckjsVersion': '2.0.5',
            'clientId': self.client_id,
        }

        self.authenticate()

        self._photos: Optional[PhotosService] = None

    def authenticate(self, force_refresh:bool=False, service:Optional[Any]=None) -> None:
        """
        Handles authentication, and persists cookies so that
        subsequent logins will not cause additional e-mails from Apple.
        """

        login_successful = False
        if self.session_data.get("session_token") and not force_refresh:
            LOGGER.debug("Checking session token validity")
            try:
                self.data = self._validate_token()
                login_successful = True
            except PyiCloudAPIResponseException:
                LOGGER.debug("Invalid authentication token, will log in from scratch.")

        if not login_successful and service is not None:
            app = self.data["apps"][service]
            if "canLaunchWithOneFactor" in app and app["canLaunchWithOneFactor"]:
                LOGGER.debug(
                    "Authenticating as %s for %s", self.user["accountName"], service
                )
                try:
                    self._authenticate_with_credentials_service(service)
                    login_successful = True
                except Exception:
                    LOGGER.debug(
                        "Could not log into service. Attempting brand new login."
                    )

        if not login_successful:
            LOGGER.debug("Authenticating as %s", self.user["accountName"])
            try:
                self._authenticate_srp()
            except PyiCloudFailedLoginException as error:
                LOGGER.error("Failed to login with srp, falling back to old raw password authentication. Error: %s", error)
                try:
                    self._authenticate_raw_password()
                except PyiCloudFailedLoginException as error:
                    LOGGER.error("Failed to login with raw password. Error: %s", error)
                    raise error

            self._authenticate_with_token()

        # Is this needed?
        self.params.update({'dsid': self.data['dsInfo']['dsid']})

        self._webservices = self.data["webservices"]

        LOGGER.info("Authentication completed successfully")
        LOGGER.debug(self.params)

    def _authenticate_with_token(self) -> None:
        """Authenticate using session token."""
        data = {
            "accountCountryCode": self.session_data.get("account_country"),
            "dsWebAuthToken": self.session_data.get("session_token"),
            "extended_login": True,
            "trustToken": self.session_data.get("trust_token", ""),
        }

        try:
            req = self.session.post(
                "%s/accountLogin" % self.SETUP_ENDPOINT, data=json.dumps(data)
            )
            self.data = req.json()
        except PyiCloudAPIResponseException as error:
            msg = "Invalid authentication token."
            raise PyiCloudFailedLoginException(msg, error) from error

        # {'domainToUse': 'iCloud.com'}
        domain_to_use = self.data.get('domainToUse')
        if domain_to_use != None:
            msg = f'Apple insists on using {domain_to_use} for your request. Please use --domain parameter'
            raise PyiCloudConnectionException(msg)

    def _authenticate_with_credentials_service(self, service: str) -> None:
        """Authenticate to a specific service using credentials."""
        data = {
            "appName": service,
            "apple_id": self.user["accountName"],
            "password": self.user["password"],
        }

        try:
            self.session.post(
                "%s/accountLogin" % self.SETUP_ENDPOINT, data=json.dumps(data)
            )

            self.data = self._validate_token()
        except PyiCloudAPIResponseException as error:
            msg = "Invalid email/password combination."
            raise PyiCloudFailedLoginException(msg, error) from error

    def _authenticate_srp(self) -> None:
        class SrpPassword():
            # srp uses the encoded password at process_challenge(), thus set_encrypt_info() should be called before that
            def __init__(self, password: str):
                self.pwd = password

            def set_encrypt_info(self, protocol: str, salt: bytes, iterations: int) -> None:
                self.protocol = protocol
                self.salt = salt
                self.iterations = iterations

            def encode(self) -> bytes:
                password_hash = hashlib.sha256(self.pwd.encode())
                password_digest = password_hash.hexdigest().encode() if self.protocol == 's2k_fo' else password_hash.digest()
                key_length = 32
                return hashlib.pbkdf2_hmac('sha256', password_digest, self.salt, self.iterations, key_length)

        # Step 1: client generates private key a (stored in srp.User) and public key A, sends to server
        srp_password = SrpPassword(self.user["password"])
        srp.rfc5054_enable()
        srp.no_username_in_x()
        usr = srp.User(self.user["accountName"], srp_password, hash_alg=srp.SHA256, ng_type=srp.NG_2048)
        uname, A = usr.start_authentication()
        data = {
            'a': base64.b64encode(A).decode(),
            'accountName': uname,
            'protocols': ['s2k', 's2k_fo']
        }

        headers = self._get_auth_headers()
        try:
            response = self.session.post("%s/signin/init" % self.AUTH_ENDPOINT, data=json.dumps(data), headers=headers)
            if response.status_code == 401:
                raise PyiCloudAPIResponseException(response.text, str(response.status_code))
        except PyiCloudAPIResponseException as error:
            msg = "Failed to initiate srp authentication."
            raise PyiCloudFailedLoginException(msg, error) from error

        # Step 2: server sends public key B, salt, and c to client
        body = response.json()
        salt = base64.b64decode(body['salt'])
        b = base64.b64decode(body['b'])
        c = body['c']
        iterations = body['iteration']
        protocol = body['protocol']

        # Step 3: client generates session key M1 and M2 with salt and b, sends to server
        srp_password.set_encrypt_info(protocol, salt, iterations)
        m1 = usr.process_challenge( salt, b )
        m2 = usr.H_AMK

        data = {
            "accountName": uname,
            "c": c,
            "m1": base64.b64encode(m1).decode(),
            "m2": base64.b64encode(m2).decode(),
            "rememberMe": True,
            "trustTokens": [],
        }

        if self.session_data.get("trust_token"):
            data["trustTokens"] = [self.session_data.get("trust_token")]

        try:
            response = self.session.post(
                "%s/signin/complete" % self.AUTH_ENDPOINT,
                params={"isRememberMeEnabled": "true"},
                data=json.dumps(data),
                headers=headers,
            )
            if response.status_code == 409:
                # requires 2FA
                pass
            elif response.status_code == 412:
                # non 2FA account returns 412 "precondition no met"
                headers = self._get_auth_headers()
                response = self.session.post(
                    "%s/repair/complete" % self.AUTH_ENDPOINT,
                    data=json.dumps({}),
                    headers=headers,
                )
            elif response.status_code >= 400 and response.status_code < 600:
                raise PyiCloudAPIResponseException(response.text, str(response.status_code))
        except PyiCloudAPIResponseException as error:
            msg = "Invalid email/password combination."
            raise PyiCloudFailedLoginException(msg, error) from error

    def _authenticate_raw_password(self) -> None:
        data = dict(self.user)

        data["rememberMe"] = True
        data["trustTokens"] = []
        if self.session_data.get("trust_token"):
            data["trustTokens"] = [self.session_data.get("trust_token")]

        headers = self._get_auth_headers()
        try:
            self.session.post(
                "%s/signin" % self.AUTH_ENDPOINT,
                params={"isRememberMeEnabled": "true"},
                data=json.dumps(data),
                headers=headers,
            )
        except PyiCloudAPIResponseException as error:
            msg = "Invalid email/password combination."
            raise PyiCloudFailedLoginException(msg, error) from error

    def _validate_token(self) -> Dict[str, Any]:
        """Checks if the current access token is still valid."""
        LOGGER.debug("Checking session token validity")
        try:
            req = self.session.post("%s/validate" % self.SETUP_ENDPOINT, data="null")
            LOGGER.debug("Session token is still valid")
            result: Dict[str, Any] = req.json()
            return result
        except PyiCloudAPIResponseException as err:
            LOGGER.debug("Invalid authentication token")
            raise err

    def _get_auth_headers(self, overrides: Optional[Dict[str, str]]=None) -> Dict[str, str]:
        headers = {
            "Accept": "application/json, text/javascript",
            "Content-Type": "application/json",
            "X-Apple-OAuth-Client-Id": "d39ba9916b7251055b22c7f910e2ea796ee65e98b2ddecea8f5dde8d9d1a815d",
            "X-Apple-OAuth-Client-Type": "firstPartyAuth",
            "X-Apple-OAuth-Redirect-URI": "https://www.icloud.com",
            "X-Apple-OAuth-Require-Grant-Code": "true",
            "X-Apple-OAuth-Response-Mode": "web_message",
            "X-Apple-OAuth-Response-Type": "code",
            "X-Apple-OAuth-State": self.client_id,
            "X-Apple-Widget-Key": "d39ba9916b7251055b22c7f910e2ea796ee65e98b2ddecea8f5dde8d9d1a815d",
        }
        scnt = self.session_data.get("scnt")
        if scnt:
            headers["scnt"] = scnt

        session_id = self.session_data.get("session_id")
        if session_id:
            headers["X-Apple-ID-Session-Id"] = session_id

        if overrides:
            headers.update(overrides)
        return headers

    @property
    def cookiejar_path(self) -> str:
        """Get path for cookiejar file."""
        return path.join(
            self._cookie_directory,
            "".join([c for c in self.user.get("accountName") if match(r"\w", c)]), # type: ignore[union-attr]
        )

    @property
    def session_path(self) -> str:
        """Get path for session data file."""
        return path.join(
            self._cookie_directory,
            "".join([c for c in self.user.get("accountName") if match(r"\w", c)]) # type: ignore[union-attr]
            + ".session",
        )

    @property
    def requires_2sa(self) -> bool:
        """Returns True if two-step authentication is required."""
        return self.data.get("dsInfo", {}).get("hsaVersion", 0) >= 1 and (
            self.data.get("hsaChallengeRequired", False) or not self.is_trusted_session
        )

    @property
    def requires_2fa(self) -> bool:
        """Returns True if two-factor authentication is required."""
        return self.data["dsInfo"].get("hsaVersion", 0) == 2 and (
            self.data.get("hsaChallengeRequired", False) or not self.is_trusted_session
        ) and self.data["dsInfo"].get("hasICloudQualifyingDevice", False)

    @property
    def is_trusted_session(self) -> bool:
        """Returns True if the session is trusted."""
        return typing.cast(bool, self.data.get("hsaTrustedBrowser", False))

    @property
    def trusted_devices(self) -> Sequence[Dict[str, Any]]:
        """ Returns devices trusted for two-step authentication."""
        request = self.session.get(
            '%s/listDevices' % self.SETUP_ENDPOINT,
            params=self.params
        )
        devices: Optional[Sequence[Dict[str, Any]]] = request.json().get('devices')
        if devices:
            return devices
        return []

    def send_request(self, request: PreparedRequest) -> Response:
        return self.session.send(request)

    def get_oauth_session(self) -> AuthenticatedSession:
        return AuthenticatedSession(client_id = self.client_id, scnt = self.session_data["scnt"], session_id = self.session_data["session_id"])

    def get_trusted_phone_numbers(self) -> Sequence[TrustedDevice]:
        """ Returns list of trusted phone number for sms 2fa """

        oauth_session = self.get_oauth_session()
        context = TrustedPhoneContextProvider(domain = self.domain, oauth_session=oauth_session)

        req = build_trusted_phone_numbers_request(context)
        request = Request(
            method = req.method,
            url = req.url,
            headers= req.headers
        ).prepare()

        response = self.send_request(request)
        
        return parse_trusted_phone_numbers_response(response)

    def send_2fa_code_sms(self, device_id: int) -> bool:
        """ Requests that a verification code is sent to the given device"""

        oauth_session = self.get_oauth_session()
        context = TrustedPhoneContextProvider(domain = self.domain, oauth_session=oauth_session)

        req = build_send_sms_code_request(context, device_id)
        request = Request(
            method = req.method,
            url = req.url,
            headers= req.headers,
            data = req.data,
            json = req.json,
        ).prepare()

        response = self.send_request(request)
        
        return response.ok

    def send_verification_code(self, device: Dict[str, Any]) -> bool:
        """ Requests that a verification code is sent to the given device"""
        data = json.dumps(device)
        request = self.session.post(
            '%s/sendVerificationCode' % self.SETUP_ENDPOINT,
            params=self.params,
            data=data
        )
        return typing.cast(bool, request.json().get('success', False))

    def validate_verification_code(self, device: Dict[str, Any], code: str) -> bool:
        """Verifies a verification code received on a trusted device."""
        device.update({"verificationCode": code, "trustBrowser": True})
        data = json.dumps(device)

        try:
            self.session.post(
                "%s/validateVerificationCode" % self.SETUP_ENDPOINT,
                params=self.params,
                data=data,
            )
        except PyiCloudAPIResponseException as error:
            if str(error.code) == "-21669":
                # Wrong verification code
                return False
            raise

        # When validating a code through the 2SA endpoint
        # the /2sv/trust URL returns 404 and won't return a trust token
        # self.trust_session()
        self._authenticate_with_token()

        return not self.requires_2sa

    def validate_2fa_code_sms(self, device_id: int, code:str) -> bool:
        """Verifies a verification code received via Apple's 2FA system through SMS."""

        oauth_session = self.get_oauth_session()
        context = TrustedPhoneContextProvider(domain = self.domain, oauth_session=oauth_session)

        req = build_verify_sms_code_request(context, device_id, code)
        request = Request(
            method = req.method,
            url = req.url,
            headers= req.headers,
            data = req.data,
            json = req.json,
        ).prepare()
        response = self.send_request(request)
        
        if response.ok:
            return self.trust_session()
        return False


    def validate_2fa_code(self, code:str) -> bool:
        """Verifies a verification code received via Apple's 2FA system (HSA2)."""
        data = {"securityCode": {"code": code}}

        headers = self._get_auth_headers({"Accept": "application/json"})

        try:
            self.session.post(
                "%s/verify/trusteddevice/securitycode" % self.AUTH_ENDPOINT,
                data=json.dumps(data),
                headers=headers,
            )
        except PyiCloudAPIResponseException as error:
            if str(error.code) == "-21669":
                # Wrong verification code
                LOGGER.error("Code verification failed.")
                return False
            raise

        LOGGER.debug("Code verification successful.")

        self.trust_session()
        return not self.requires_2sa

    def trust_session(self) -> bool:
        """Request session trust to avoid user log in going forward."""
        headers = self._get_auth_headers()

        try:
            self.session.get(
                f"{self.AUTH_ENDPOINT}/2sv/trust",
                headers=headers,
            )
            self._authenticate_with_token()
            return True
        except PyiCloudAPIResponseException:
            LOGGER.error("Session trust failed.")
            return False

    def _get_webservice_url(self, ws_key: str) -> str:
        """Get webservice URL, raise an exception if not exists."""
        if self._webservices.get(ws_key) is None:
            raise PyiCloudServiceNotActivatedException(
                "Webservice not available", ws_key
            )
        return typing.cast(str, self._webservices[ws_key]["url"])

    @property
    def devices(self) -> Sequence[AppleDevice]: 
        """ Return all devices."""
        service_root = self._get_webservice_url("findme")
        return typing.cast(Sequence[AppleDevice], FindMyiPhoneServiceManager(
            service_root,
            self.session,
            self.params
        ))

    @property
    def account(self): # type: ignore
        service_root = self._gget_webservice_url("account") # type: ignore
        return AccountService( # type: ignore
            service_root,
            self.session,
            self.params
        )

    @property
    def iphone(self): # type: ignore
        return self.devices[0]

    @property
    def files(self): # type: ignore
        if not hasattr(self, '_files'):
            service_root = self._get_webservice_url("ubiquity")
            self._files = UbiquityService( # type: ignore
                service_root,
                self.session,
                self.params
            )
        return self._files

    @property
    def photos(self) -> PhotosService:
        """Gets the 'Photo' service."""
        if not self._photos:
            service_root = self._get_webservice_url("ckdatabasews")
            self._photos = PhotosService(
                service_root, 
                self.session, 
                self.params, 
                self.filename_cleaner, 
                self.lp_filename_generator, 
                self.raw_policy,
                self.file_match_policy
                )
        return self._photos

    @property
    def calendar(self): # type: ignore
        service_root = self._get_webservice_url("calendar")
        return CalendarService(service_root, self.session, self.params)# type: ignore

    @property
    def contacts(self): # type: ignore
        service_root = self._get_webservice_url("contacts")
        return ContactsService(service_root, self.session, self.params)# type: ignore

    @property
    def reminders(self): # type: ignore
        service_root = self._get_webservice_url("reminders")
        return RemindersService(service_root, self.session, self.params)# type: ignore

    def __unicode__(self) -> str:
        return 'iCloud API: %s' % self.user.get('accountName')

    def __str__(self) -> str:
        as_unicode = self.__unicode__()
        if sys.version_info[0] >= 3:
            return as_unicode
        else:
            return as_unicode.encode('ascii', 'ignore')

    def __repr__(self) -> str:
        return '<%s>' % str(self)



import os
import sys
from logging import Logger
from typing import Union

import waitress
from flask import Flask, Response, make_response, render_template, request

from icloudpd.status import Status, StatusExchange


def serve_app(logger: Logger, _status_exchange: StatusExchange) -> None:
    app = Flask(__name__)
    app.logger = logger
    # for running in pyinstaller
    bundle_dir = getattr(sys, "_MEIPASS", None)
    if bundle_dir is not None:
        app.template_folder = os.path.join(bundle_dir, "templates")
        app.static_folder = os.path.join(bundle_dir, "static")

    @app.route("/")
    def index() -> Union[Response, str]:
        return render_template("index.html")

    @app.route("/status", methods=["GET"])
    def get_status() -> Union[Response, str]:
        _status = _status_exchange.get_status()
        _config = _status_exchange.get_config()
        _progress = _status_exchange.get_progress()
        if _status == Status.NO_INPUT_NEEDED:
            return render_template(
                "no_input.html", status=_status, progress=_progress, config=vars(_config)
            )
        if _status == Status.NEED_MFA:
            return render_template("code.html")
        if _status == Status.NEED_PASSWORD:
            return render_template("password.html", config=_config)
        return render_template("status.html", status=_status)

    @app.route("/code", methods=["POST"])
    def set_code() -> Union[Response, str]:
        _config = _status_exchange.get_config()
        code = request.form.get("code")
        if code is not None:
            if _status_exchange.set_payload(code):
                return render_template("code_submitted.html")
        else:
            logger.error(f"cannot find code in request {request.form}")
        return make_response(
            render_template("auth_error.html", config=_config, type="Two-Factor Code"), 400
        )  # incorrect code

    @app.route("/password", methods=["POST"])
    def set_password() -> Union[Response, str]:
        _config = _status_exchange.get_config()
        password = request.form.get("password")
        if password is not None:
            if _status_exchange.set_payload(password):
                return render_template("password_submitted.html")
        else:
            logger.error(f"cannot find password in request {request.form}")
        return make_response(
            render_template("auth_error.html", config=_config, type="password"), 400
        )  # incorrect code

    @app.route("/resume", methods=["POST"])
    def resume() -> Union[Response, str]:
        _status_exchange.get_progress().resume = True
        return make_response("Ok", 200)

    @app.route("/cancel", methods=["POST"])
    def cancel() -> Union[Response, str]:
        _status_exchange.get_progress().cancel = True
        return make_response("Ok", 200)

    logger.debug("Starting web server...")
    return waitress.serve(app)



"""
Delete any files found in "Recently Deleted"
"""

import datetime
import logging
import os
from typing import Sequence, Set

from tzlocal import get_localzone

from icloudpd.paths import local_download_path
from pyicloud_ipd.services.photos import PhotoLibrary
from pyicloud_ipd.utils import disambiguate_filenames
from pyicloud_ipd.version_size import AssetVersionSize, VersionSize


def delete_file(logger: logging.Logger, path: str) -> bool:
    """Actual deletion of files"""
    os.remove(path)
    logger.info("Deleted %s", path)
    return True


def delete_file_dry_run(logger: logging.Logger, path: str) -> bool:
    """Dry run deletion of files"""
    logger.info("[DRY RUN] Would delete %s", path)
    return True


def autodelete_photos(
    logger: logging.Logger,
    dry_run: bool,
    library_object: PhotoLibrary,
    folder_structure: str,
    directory: str,
    _sizes: Sequence[AssetVersionSize],
) -> None:
    """
    Scans the "Recently Deleted" folder and deletes any matching files
    from the download directory.
    (I.e. If you delete a photo on your phone, it's also deleted on your computer.)
    """
    logger.info("Deleting any files found in 'Recently Deleted'...")

    recently_deleted = library_object.albums["Recently Deleted"]

    for media in recently_deleted:
        try:
            created_date = media.created.astimezone(get_localzone())
        except (ValueError, OSError):
            logger.error("Could not convert media created date to local timezone %s", media.created)
            created_date = media.created

        if folder_structure.lower() == "none":
            date_path = ""
        else:
            try:
                date_path = folder_structure.format(created_date)
            except ValueError:  # pragma: no cover
                # This error only seems to happen in Python 2
                logger.error("Photo created date was not valid (%s)", created_date)
                # e.g. ValueError: year=5 is before 1900
                # (https://github.com/icloud-photos-downloader/icloud_photos_downloader/issues/122)
                # Just use the Unix epoch
                created_date = datetime.datetime.fromtimestamp(0)
                date_path = folder_structure.format(created_date)

        download_dir = os.path.join(directory, date_path)

        paths: Set[str] = set({})
        _size: VersionSize
        for _size, _version in disambiguate_filenames(media.versions, _sizes).items():
            if _size in [AssetVersionSize.ALTERNATIVE, AssetVersionSize.ADJUSTED]:
                paths.add(os.path.normpath(local_download_path(_version.filename, download_dir)))
                paths.add(
                    os.path.normpath(local_download_path(_version.filename, download_dir)) + ".xmp"
                )
        for _size, _version in media.versions.items():
            if _size not in [AssetVersionSize.ALTERNATIVE, AssetVersionSize.ADJUSTED]:
                paths.add(os.path.normpath(local_download_path(_version.filename, download_dir)))
                paths.add(
                    os.path.normpath(local_download_path(_version.filename, download_dir)) + ".xmp"
                )
        for path in paths:
            if os.path.exists(path):
                logger.debug("Deleting %s...", path)
                delete_local = delete_file_dry_run if dry_run else delete_file
                delete_local(logger, path)



"""Path functions"""

import os


def remove_unicode_chars(value: str) -> str:
    """Removes unicode chars from the string"""
    result = value.encode("utf-8").decode("ascii", "ignore")
    return result


def clean_filename(filename: str) -> str:
    """Replaces invalid chars in filenames with '_'"""
    invalid = '<>:"/\\|?*\0'
    result = filename

    for char in invalid:
        result = result.replace(char, "_")

    return result


def local_download_path(filename: str, download_dir: str) -> str:
    """Returns the full download path, including size"""
    download_path = os.path.join(download_dir, filename)
    return download_path



from enum import Enum
from threading import Lock
from typing import Optional

from icloudpd.config import Config
from icloudpd.progress import Progress


class Status(Enum):
    NO_INPUT_NEEDED = "no_input_needed"
    NEED_MFA = "need_mfa"
    SUPPLIED_MFA = "supplied_mfa"
    CHECKING_MFA = "checking_mfa"
    NEED_PASSWORD = "need_password"
    SUPPLIED_PASSWORD = "supplied_password"
    CHECKING_PASSWORD = "checking_password"

    def __str__(self) -> str:
        return self.name


class StatusExchange:
    def __init__(self) -> None:
        self.lock = Lock()
        self._status = Status.NO_INPUT_NEEDED
        self._payload: Optional[str] = None
        self._config: Optional[Config] = None
        self._progress = Progress()

    def get_status(self) -> Status:
        with self.lock:
            return self._status

    def replace_status(self, expected_status: Status, new_status: Status) -> bool:
        with self.lock:
            if self._status == expected_status:
                self._status = new_status
                return True
            else:
                return False

    def set_payload(self, payload: str) -> bool:
        with self.lock:
            if self._status != Status.NEED_MFA and self._status != Status.NEED_PASSWORD:
                return False

            self._payload = payload
            self._status = (
                Status.SUPPLIED_MFA if self._status == Status.NEED_MFA else Status.SUPPLIED_PASSWORD
            )
            return True

    def get_payload(self) -> Optional[str]:
        with self.lock:
            if self._status not in [
                Status.SUPPLIED_MFA,
                Status.CHECKING_MFA,
                Status.SUPPLIED_PASSWORD,
                Status.CHECKING_PASSWORD,
            ]:
                return None

            return self._payload

    def set_config(self, config: Config) -> None:
        with self.lock:
            self._config = config

    def get_config(self) -> Optional[Config]:
        with self.lock:
            return self._config

    def get_progress(self) -> Progress:
        with self.lock:
            return self._progress



from pyicloud_ipd.services.photos import PhotoAsset

from typing import Iterable
from icloudpd.counter import Counter
from icloudpd.download import mkdirs_for_path

import itertools
import logging
import os


def handle_recent_until_found(
    photos_count: int | None,
    photos_enumerator: Iterable[PhotoAsset],
    recent: int | None,
    until_found: int | None,
) -> tuple[int | None, Iterable[PhotoAsset]]:
    if recent is not None:
        photos_count = recent
        photos_enumerator = itertools.islice(photos_enumerator, recent)

    if until_found is not None:
        photos_count = None
        # ensure photos iterator doesn't have a known length
        photos_enumerator = (p for p in photos_enumerator)

    return photos_count, iter(photos_enumerator)


def log_at_download_start(
    logger: logging.Logger,
    photos_count: int | None,
    primary_sizes: list[str],
    skip_videos: bool,
    directory: str,
) -> None:
    if photos_count is not None:
        plural_suffix = "" if photos_count == 1 else "s"
        video_suffix = ""
        photos_count_str = "the first" if photos_count == 1 else photos_count

        if not skip_videos:
            video_suffix = " or video" if photos_count == 1 else " and videos"
    else:
        photos_count_str = "???"
        plural_suffix = "s"
        video_suffix = " and videos" if not skip_videos else ""
    logger.info(
        ("Downloading %s %s" + " photo%s%s to %s ..."),
        photos_count_str,
        ",".join(primary_sizes),
        plural_suffix,
        video_suffix,
        directory,
    )


def should_break(counter: Counter, until_found: int | None) -> bool:
    """Exit if until_found condition is reached"""
    return until_found is not None and counter.value() >= until_found


def check_folder_structure(logger: logging.Logger, directory: str, folder_structure: str) -> None:
    """
    Check if there exists a .folderstructure file in the directory. If not, create it.
    If the file exists, check if the folder structure is the same as the one in the file.

    Return if the folder structure is the same or the folder is newly created.
    Raise an error if the folder structure is different or there are files in the directory without
    the folder structure.

    Note that this check cannot prevent the user from altering the structure file manually with the
    .folderstructure file in place.
    """

    def write_structure_file(structure_file_path: str, folder_structure: str) -> None:
        with open(structure_file_path, "w") as f:
            logger.info(
                f"Creating .folderstructure file in {directory} with folder structure: {folder_structure}"
            )
            f.write(folder_structure + "\n")
        os.chmod(structure_file_path, 0o644)

    structure_file_path = os.path.join(directory, ".folderstructure")

    # folder does not exist
    if not os.path.exists(directory):
        mkdirs_for_path(logger, structure_file_path)
        write_structure_file(structure_file_path, folder_structure)
        return

    directory_empty = not [f for f in os.listdir(directory) if not f.startswith(".")]

    if directory_empty:
        write_structure_file(structure_file_path, folder_structure)
        return

    # folder not empty but no .structure file
    if not directory_empty and not os.path.exists(structure_file_path):
        raise ValueError(
            "Cannot determine the structure of a non-empty directory. Please provide a .folderstructure file manually or use an empty directory."
        )

    # folder exists and .structure file exists
    with open(structure_file_path, "r") as f:
        if (provided_structure := f.read().strip()) != folder_structure:
            raise ValueError(
                f"The specified folder structure: {folder_structure} is different from the one found in the existing .folderstructure file: {provided_structure}"
            )
        else:
            logger.info(
                f"Continue downloading to {directory} with the folder structure: {folder_structure}"
            )



from typing import Callable, Tuple, TypeVar
from . import optional
_T_contra = TypeVar("_T_contra", contravariant=True)
_T2_contra = TypeVar("_T2_contra", contravariant=True)
_T3_contra = TypeVar("_T3_contra", contravariant=True)
_T_co = TypeVar("_T_co", covariant=True)
_T_inv = TypeVar("_T_inv")


def compose(
    f: Callable[[_T_inv], _T_co], g: Callable[[_T_contra], _T_inv]
) -> Callable[[_T_contra], _T_co]:
    """
    `f after g` composition of functions

    Equiv: lamdba x -> f(g(x))
    """

    def inter_(value: _T_contra) -> _T_co:
        return f(g(value))

    return inter_


def identity(value: _T_inv) -> _T_inv:
    """
    identity function
    """
    return value


def constant(value: _T_inv) -> Callable[[_T_contra], _T_inv]:
    """
    constant function
    """

    def _intern(_: _T_contra) -> _T_inv:
        return value

    return _intern


def pipe(
    f: Callable[[_T_contra], _T_inv], g: Callable[[_T_inv], _T_co]
) -> Callable[[_T_contra], _T_co]:
    """
    `g after f` composition of functions (reverse of compose)
    """
    return compose(g, f)


def apply_reverse(input: _T_contra) -> Callable[[Callable[[_T_contra], _T_co]], _T_co]:
    """
    Applying a function. Equiv curried `(&)` in Haskel
    a -> (a -> b) -> b

    Example usage:

        >>> def _mul(a: int, b: int) -> int:
        ...     return a * b
        >>> def _add(a: int, b: int) -> int:
        ...     return a + b
        >>> list(map(apply_reverse(3), [curry2(_mul)(2), curry2(_add)(5)])) == [6, 8]
        True

    """

    def _intern(func: Callable[[_T_contra], _T_co]) -> _T_co:
        return func(input)

    return _intern


def curry2(
    func: Callable[[_T_contra, _T2_contra], _T_co],
) -> Callable[[_T_contra], Callable[[_T2_contra], _T_co]]:
    """
    Transforms 2-param function into two nested 1-param functions

    >>> def _mul(a: int, b: int) -> int:
    ...     return a * b
    >>> _mul(2, 3) == 6
    True
    >>> curry2(_mul)(2)(3) == 6
    True

    """

    def _intern(input: _T_contra) -> Callable[[_T2_contra], _T_co]:
        def _intern2(input2: _T2_contra) -> _T_co:
            return func(input, input2)

        return _intern2

    return _intern


def uncurry2(
    func: Callable[[_T_contra], Callable[[_T2_contra], _T_co]],
) -> Callable[[_T_contra, _T2_contra], _T_co]:
    """
    Transforms two nested 1-param functions into one 2-param function

    >>> def _mul_c(a: int):
    ...     def _intern(b: int) -> int:
    ...         return a * b
    ...
    ...     return _intern
    >>> _mul_c(2)(3) == 6
    True
    >>> uncurry2(_mul_c)(2, 3) == 6
    True

    """

    def _intern(input: _T_contra, input2: _T2_contra) -> _T_co:
        return func(input)(input2)

    return _intern


def curry3(
    func: Callable[[_T_contra, _T2_contra, _T3_contra], _T_co],
) -> Callable[[_T_contra], Callable[[_T2_contra], Callable[[_T3_contra], _T_co]]]:
    def _intern(input: _T_contra) -> Callable[[_T2_contra], Callable[[_T3_contra], _T_co]]:
        def _intern2(input2: _T2_contra) -> Callable[[_T3_contra], _T_co]:
            def _intern3(input3: _T3_contra) -> _T_co:
                return func(input, input2, input3)

            return _intern3

        return _intern2

    return _intern


def fst(t: Tuple[_T_inv, _T2_contra]) -> _T_inv:
    """get first of tuple
    >>> fst((1, 2)) == 1
    True
    """
    return t[0]


def snd(t: Tuple[_T_contra, _T_inv]) -> _T_inv:
    """get second of tuple
    >>> snd((1, 2)) == 2
    True
    """
    return t[1]


def flip(
    func: Callable[[_T_contra, _T2_contra], _T_co],
) -> Callable[[_T2_contra, _T_contra], _T_co]:
    """flips params
    >>> def minus(a: int, b: int) -> int:
    ...     return a - b
    >>> minus(5, 3) == 2
    True
    >>> flip(minus)(5, 3) == -2
    True
    """

    def _intern(input2: _T2_contra, input: _T_contra) -> _T_co:
        return func(input, input2)

    return _intern


def compact2(
    func: Callable[[_T_contra, _T2_contra], _T_co],
) -> Callable[[Tuple[_T_contra, _T2_contra]], _T_co]:
    """Compacts two parameters into one tuple"""

    def _intern(input: Tuple[_T_contra, _T2_contra]) -> _T_co:
        return func(fst(input), snd(input))

    return _intern


def expand2(
    func: Callable[[Tuple[_T_contra, _T2_contra]], _T_co],
) -> Callable[[_T_contra, _T2_contra], _T_co]:
    def _intern(input: _T_contra, input2: _T2_contra) -> _T_co:
        return func((input, input2))

    return _intern


def pipe2(
    f: Callable[[_T_contra, _T2_contra], _T_inv], g: Callable[[_T_inv], _T_co]
) -> Callable[[_T_contra, _T2_contra], _T_co]:
    """
    `g after f` composition of functions (reverse of compose). f takes 2 params
    >>> def mul(a: int, b: int) -> int:
    ...     return a * b
    >>> def add5(a: int) -> int:
    ...     return a + 5
    >>> pipe2(mul, add5)(2, 3) == 11
    True
    """
    return expand2(pipe(compact2(f), g))

__all__ = [
    'compose',
    'identity',
    'constant',
    'pipe',
    'apply_reverse',
    'curry2',
    'uncurry2',
    'curry3',
    'fst',
    'snd',
    'flip',
    'compact2',
    'expand2',
    'pipe2',
]


from pyicloud_ipd.services.calendar import CalendarService
from pyicloud_ipd.services.findmyiphone import FindMyiPhoneServiceManager
from pyicloud_ipd.services.ubiquity import UbiquityService
from pyicloud_ipd.services.contacts import ContactsService
from pyicloud_ipd.services.reminders import RemindersService
from pyicloud_ipd.services.photos import PhotosService
from pyicloud_ipd.services.account import AccountService
